import asyncio

from rasa_sdk import Action
from rasa_sdk.events import EventType
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.interfaces import Tracker
from rasa_sdk.types import DomainDict


class ActionSleepAndRespond(Action):
    def name(self) -> str:
        return "action_sleep_few_sec"

    async def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: DomainDict
    ) -> list[EventType]:
        await asyncio.sleep(3)
        return []



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict

import structlog

from rasa.dialogue_understanding.stack.frames import (
    PatternFlowStackFrame,
)
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX

structlogger = structlog.get_logger()

FLOW_PATTERN_CODE_CHANGE_ID = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "code_change"


@dataclass
class CodeChangeFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame which cleans the stack after a bot update."""

    flow_id: str = FLOW_PATTERN_CODE_CHANGE_ID
    """The ID of the flow."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_CODE_CHANGE_ID

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> CodeChangeFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return CodeChangeFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
        )



import argparse
import base64
import re
import sys
from typing import Any, Dict, Iterable, List, Optional, Set, Text, Tuple, Union

import questionary
import requests
import structlog

import rasa.cli.telemetry
import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.shared.constants import (
    CONFIG_LANGUAGE_KEY,
    CONFIG_LLM_KEY,
    CONFIG_MODEL_NAME_KEY,
    CONFIG_PIPELINE_KEY,
    CONFIG_POLICIES_KEY,
    CONFIG_RECIPE_KEY,
    DEFAULT_CONFIG_PATH,
    DEFAULT_DOMAIN_PATHS,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows.yaml_flows_io import YAMLFlowsReader
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.importers.utils import (
    CALMUserData,
    extract_calm_import_parts_from_importer,
)
from rasa.shared.nlu.training_data.formats.rasa_yaml import (
    RasaYAMLReader,
    RasaYAMLWriter,
)
from rasa.shared.utils.llm import collect_custom_prompts
from rasa.shared.utils.yaml import (
    dump_obj_as_yaml_to_string,
    read_yaml_file,
)
from rasa.studio import results_logger
from rasa.studio.auth import KeycloakTokenReader
from rasa.studio.config import StudioConfig
from rasa.studio.results_logger import StudioResult, with_studio_error_handler
from rasa.studio.utils import validate_argument_paths
from rasa.telemetry import track_upload_to_studio_failed
from rasa.utils.json_utils import extract_values

structlogger = structlog.get_logger()

CONFIG_KEYS = [
    "recipe",
    "language",
    "pipeline",
    "llm",
    "policies",
    "model_name",
    "assistant_id",
]

DOMAIN_KEYS = [
    "version",
    "actions",
    "responses",
    "slots",
    "intents",
    "entities",
    "forms",
    "session_config",
]


def _get_selected_entities_and_intents(
    args: argparse.Namespace,
    intents_from_files: Set[Text],
    entities_from_files: List[Text],
) -> Tuple[List[Text], List[Text]]:
    entities = args.entities

    if entities is None or len(entities) == 0:
        entities = entities_from_files
        structlogger.info(
            "rasa.studio.upload.entities_empty",
            event_info="No entities specified. Using all entities from files.",
        )

    intents = args.intents

    if intents is None or len(intents) == 0:
        intents = intents_from_files
        structlogger.info(
            "rasa.studio.upload.intents_empty",
            event_info="No intents specified. Using all intents from files.",
        )

    return list(entities), list(intents)


def run_validation(args: argparse.Namespace) -> None:
    """Run the validation before uploading to Studio.

    This is to avoid uploading invalid assistant data
    that would raise errors during Rasa Pro training in Studio.

    The validation checks that were selected to be run before uploading
    maintain parity with the features that are supported in Studio.
    """
    from rasa.validator import Validator

    training_data_paths = args.data if isinstance(args.data, list) else [args.data]
    training_data_importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain,
        training_data_paths=training_data_paths,
        config_path=args.config,
        expand_env_vars=False,
    )

    structlogger.info(
        "rasa.studio.upload.validating_data",
        event_info="Validating domain and training data...",
    )

    validator = Validator.from_importer(training_data_importer)

    if not validator.verify_studio_supported_validations():
        structlogger.error(
            "rasa.studio.upload.validate_files.project_validation_error",
            event_info="Project validation completed with errors.",
        )
        sys.exit(1)

    structlogger.info(
        "rasa.studio.upload.validate_files.success",
        event_info="Project validation completed successfully.",
    )


def handle_upload(args: argparse.Namespace) -> None:
    """Uploads primitives to rasa studio."""
    validate_argument_paths(args)
    studio_config = StudioConfig.read_config()
    endpoint = studio_config.studio_url
    verify = not studio_config.disable_verify

    if not endpoint:
        rasa.shared.utils.cli.print_error_and_exit(
            "No GraphQL endpoint found in config. Please run `rasa studio config`."
        )

    if not is_auth_working(endpoint, verify):
        rasa.shared.utils.cli.print_error_and_exit(
            "Authentication is invalid or expired. Please run `rasa studio login`."
        )

    structlogger.info("rasa.studio.upload.loading_data", event_info="Loading data...")

    args.domain = rasa.cli.utils.get_validated_path(
        args.domain, "domain", DEFAULT_DOMAIN_PATHS
    )

    args.config = rasa.cli.utils.get_validated_path(
        args.config, "config", DEFAULT_CONFIG_PATH
    )

    config = read_yaml_file(args.config, expand_env_vars=False)
    assistant_name = args.assistant_name or _get_assistant_name(config)
    args.assistant_name = assistant_name
    if not _handle_existing_assistant(
        assistant_name, studio_config.studio_url, verify, args
    ):
        return

    Domain.expand_env_vars = False
    RasaYAMLReader.expand_env_vars = False
    YAMLFlowsReader.expand_env_vars = False

    upload_calm_assistant(args, endpoint, verify=verify)


config_keys = [
    CONFIG_RECIPE_KEY,
    CONFIG_POLICIES_KEY,
    CONFIG_PIPELINE_KEY,
    CONFIG_LANGUAGE_KEY,
    CONFIG_LLM_KEY,
    CONFIG_MODEL_NAME_KEY,
]


def _get_assistant_name(config: Dict[Text, Any]) -> str:
    config_assistant_id = config.get("assistant_id", "")
    assistant_name = questionary.text(
        "Please provide the assistant name", default=config_assistant_id
    ).ask()
    if not assistant_name:
        structlogger.error(
            "rasa.studio.upload.assistant_name_empty",
            event_info="Assistant name cannot be empty.",
        )
        sys.exit(1)

    # if assistant_name exists and different from config assistant_id,
    # notify user and upload with new assistant_name
    same = assistant_name == config_assistant_id
    if not same and config_assistant_id != "":
        structlogger.info(
            "rasa.studio.upload.assistant_name_mismatch",
            event_info=(
                f"Assistant name '{assistant_name}' is different"
                f" from the one in the config file: '{config_assistant_id}'."
            ),
        )

    structlogger.info(
        "rasa.studio.upload.name_selected",
        event_info=f"Uploading assistant with the name '{assistant_name}'.",
        assistant_name=assistant_name,
    )
    return assistant_name


def build_calm_import_parts(
    data_path: Union[Text, List[Text]],
    domain_path: Text,
    config_path: Text,
    endpoints_path: Optional[Text] = None,
    assistant_name: Optional[Text] = None,
) -> Tuple[str, CALMUserData]:
    """Builds the parts of the assistant to be uploaded to Studio.

    Args:
        data_path: The path to the training data
        domain_path: The path to the domain
        config_path: The path to the config
        endpoints_path: The path to the endpoints
        assistant_name: The name of the assistant

    Returns:
        The assistant name and the parts to be uploaded
    """
    training_data_paths = data_path if isinstance(data_path, list) else [str(data_path)]
    importer = TrainingDataImporter.load_from_dict(
        domain_path=domain_path,
        config_path=config_path,
        training_data_paths=training_data_paths,
        expand_env_vars=False,
    )

    config = read_yaml_file(config_path, expand_env_vars=False)
    endpoints = read_yaml_file(endpoints_path, expand_env_vars=False)
    assistant_name = assistant_name or _get_assistant_name(config)

    parts = extract_calm_import_parts_from_importer(
        importer=importer,
        config=config,
        endpoints=endpoints,
    )

    return assistant_name, parts


@with_studio_error_handler
def upload_calm_assistant(
    args: argparse.Namespace, endpoint: str, verify: bool = True
) -> StudioResult:
    def yaml_or_empty(part: Dict[Text, Any]) -> Optional[str]:
        return dump_obj_as_yaml_to_string(part) if part else None

    run_validation(args)
    structlogger.info(
        "rasa.studio.upload.loading_data", event_info="Parsing CALM assistant data..."
    )
    assistant_name, parts = build_calm_import_parts(
        data_path=args.data,
        domain_path=args.domain,
        config_path=args.config,
        endpoints_path=args.endpoints,
        assistant_name=args.assistant_name,
    )

    prompts_json = collect_custom_prompts(parts.config, parts.endpoints)
    graphql_req = build_import_request(
        assistant_name,
        flows_yaml=yaml_or_empty(parts.flows),
        domain_yaml=yaml_or_empty(parts.domain),
        config_yaml=yaml_or_empty(parts.config),
        endpoints=yaml_or_empty(parts.endpoints),
        nlu_yaml=yaml_or_empty(parts.nlu),
        prompts_json=prompts_json,
    )

    structlogger.info(
        "rasa.studio.upload.calm", event_info="Uploading to Rasa Studio..."
    )
    return make_request(endpoint, graphql_req, verify)


@with_studio_error_handler
def upload_nlu_assistant(
    args: argparse.Namespace, endpoint: str, verify: bool = True
) -> StudioResult:
    """Uploads the classic (dm1) assistant data to Rasa Studio.

    Args:
        args: The command line arguments
            - data: The path to the training data
            - domain: The path to the domain
            - intents: The intents to upload
            - entities: The entities to upload
        endpoint: The studio endpoint
        verify: Whether to verify SSL
    Returns:
        None
    """
    structlogger.info(
        "rasa.studio.upload.nlu_data_read",
        event_info="Found DM1 assistant data, parsing...",
    )
    training_data_paths = args.data if isinstance(args.data, list) else [args.data]
    importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain,
        training_data_paths=training_data_paths,
        config_path=args.config,
        expand_env_vars=False,
    )

    intents_from_files = importer.get_nlu_data().intents

    domain_from_files = importer.get_domain()
    entities_from_files = domain_from_files.entities
    entities, intents = _get_selected_entities_and_intents(
        args, intents_from_files, entities_from_files
    )

    config_from_files = importer.get_config()
    config = extract_values(config_from_files, CONFIG_KEYS)

    assistant_name = _get_assistant_name(config)

    structlogger.info(
        "rasa.studio.upload.nlu_data_validate", event_info="Validating data..."
    )
    _check_for_missing_primitives(
        intents, entities, intents_from_files, entities_from_files
    )

    nlu_examples = importer.get_nlu_data().filter_training_examples(
        lambda ex: ex.get("intent") in intents
    )

    all_entities = _add_missing_entities(nlu_examples.entities, entities)
    nlu_examples_yaml = RasaYAMLWriter().dumps(nlu_examples)

    domain = _filter_domain(all_entities, intents, domain_from_files.as_dict())
    domain_yaml = dump_obj_as_yaml_to_string(domain)

    graphql_req = build_request(assistant_name, nlu_examples_yaml, domain_yaml)

    structlogger.info(
        "rasa.studio.upload.nlu", event_info="Uploading to Rasa Studio..."
    )
    return make_request(endpoint, graphql_req, verify)


def is_auth_working(endpoint: str, verify: bool = True) -> bool:
    """Send a test request to Studio to check if auth is working."""
    result = make_request(
        endpoint,
        {
            "operationName": "LicenseDetails",
            "query": (
                "query LicenseDetails {\n"
                "  licenseDetails {\n"
                "    valid\n"
                "    scopes\n"
                "    __typename\n"
                "  }\n"
                "}"
            ),
            "variables": {},
        },
        verify,
    )
    return result.was_successful


def make_request(endpoint: str, graphql_req: Dict, verify: bool = True) -> StudioResult:
    """Makes a request to the studio endpoint to upload data.

    Args:
        endpoint: The studio endpoint
        graphql_req: The graphql request
        verify: Whether to verify SSL
    """
    token = KeycloakTokenReader().get_token()
    res = requests.post(
        endpoint,
        json=graphql_req,
        headers={
            "Authorization": f"{token.token_type} {token.access_token}",
            "Content-Type": "application/json",
        },
        verify=verify,
    )
    if results_logger.response_has_errors(res.json()):
        track_upload_to_studio_failed(res.json())
        return StudioResult.error(res.json())
    return StudioResult.success("Upload successful")


def _add_missing_entities(
    entities_from_intents: Iterable[str], entities: List[str]
) -> List[Union[str, Dict]]:
    all_entities: List[Union[str, Dict]] = []
    all_entities.extend(entities)
    for entity in entities_from_intents:
        if entity not in entities:
            structlogger.warning(
                "rasa.studio.upload.adding_missing_entity",
                event_info=(
                    f"Adding entity '{entity}' to upload "
                    "since it is used in an intent."
                ),
                entity=entity,
            )
            all_entities.append(entity)
    return all_entities


def build_import_request(
    assistant_name: str,
    flows_yaml: Optional[str] = None,
    domain_yaml: Optional[str] = None,
    config_yaml: Optional[str] = None,
    endpoints: Optional[str] = None,
    nlu_yaml: Optional[str] = None,
    prompts_json: Optional[Dict[str, str]] = None,
) -> Dict:
    """Builds the GraphQL request for uploading a modern assistant.

    Args:
        assistant_name: The name of the assistant
        flows_yaml: The YAML representation of the flows
        domain_yaml: The YAML representation of the domain
        config_yaml: The YAML representation of the config
        endpoints: The YAML representation of the endpoints
        nlu_yaml: The YAML representation of the NLU data
        prompts_json: The JSON representation of the prompts

    Returns:
        A dictionary representing the GraphQL request for uploading the assistant.
    """
    inputs_map = {
        "domain": domain_yaml,
        "flows": flows_yaml,
        "config": config_yaml,
        "endpoints": endpoints,
        "nlu": nlu_yaml,
    }

    payload: Dict[Text, Any] = {
        field: convert_string_to_base64(value)
        for field, value in inputs_map.items()
        if value is not None
    }

    if prompts_json:
        payload["prompts"] = prompts_json

    variables_input = {"assistantName": assistant_name, **payload}

    graphql_req = {
        "query": (
            "mutation UploadModernAssistant($input: UploadModernAssistantInput!)"
            "{\n  uploadModernAssistant(input: $input)\n}"
        ),
        "variables": {"input": variables_input},
    }

    return graphql_req


def convert_string_to_base64(string: str) -> str:
    """Converts a string to base64.

    Args:
        string: The string to convert

    Returns:
        The base64 encoded string
    """
    return base64.b64encode(string.encode("utf-8")).decode("utf-8")


def build_request(
    assistant_name: str, nlu_examples_yaml: str, domain_yaml: str
) -> Dict:
    # b64encode expects bytes and returns bytes so we need to decode to string
    base64_domain = base64.b64encode(domain_yaml.encode("utf-8")).decode("utf-8")
    base64_nlu_examples = base64.b64encode(nlu_examples_yaml.encode("utf-8")).decode(
        "utf-8"
    )

    graphql_req = {
        "query": (
            "mutation ImportFromEncodedYaml($input: ImportFromEncodedYamlInput!)"
            "{\n  importFromEncodedYaml(input: $input)\n}"
        ),
        "variables": {
            "input": {
                "assistantName": assistant_name,
                "domain": base64_domain,
                "nlu": base64_nlu_examples,
            }
        },
    }

    return graphql_req


def _filter_domain(
    entities: List[Union[str, Dict]],
    intents: List[str],
    domain_from_files: Dict[str, Any],
) -> Dict:
    """Filters the domain to only include the selected entities and intents."""
    selected_entities = _remove_not_selected_entities(
        entities, domain_from_files.get("entities", [])
    )
    return {
        "version": domain_from_files["version"],
        "intents": intents,
        "entities": selected_entities,
    }


def _check_for_missing_primitives(
    intents: Iterable[str],
    entities: Iterable[str],
    intents_found: Iterable[str],
    entities_found: Iterable[str],
) -> None:
    """Checks if the data contains all intents and entities.

    Args:
        intents: Iterable of intents to check
        entities: Iterable of entities to check
        intents_found: Iterable of intents found in the data
        entities_found: Iterable of entities found in the data
    """
    not_found_entities = [entity for entity in entities if entity not in entities_found]
    not_found_intents = [intent for intent in intents if intent not in intents_found]

    not_found_msg = ""
    if not_found_entities:
        not_found_msg += (
            f"Entities were not found in the training data: {not_found_entities}\n"
        )
    if not_found_intents:
        not_found_msg += (
            f"Intents were not found in the training data: {not_found_intents}\n"
        )

    if not_found_msg:
        raise RasaException(not_found_msg)


def _remove_not_selected_entities(
    entities: List[Union[Text, Dict]], domain_entities: List[Union[Text, Dict]]
) -> List:
    to_remove: List[Union[Text, Dict]] = []

    for entity in domain_entities:
        if isinstance(entity, str) and entity not in entities:
            to_remove.append(entity)
        elif isinstance(entity, dict) and len(entity) == 1:
            entity_name = next(iter(entity))
            if entity_name not in entities:
                to_remove.append(entity)

    for entity in to_remove:
        domain_entities.remove(entity)

    return domain_entities


def remove_quotes(node: Any) -> Any:
    """Transform function to remove quotes from a node if it is a string.

    This is to prevent wrapping unexpanded environment variables in quotes
    when uploading endpoints to Rasa Studio.
    """
    if isinstance(node, str):
        matches = re.findall(r"'\$\{([^}]+)\}'", node)
        for match in matches:
            node = node.replace(f"'${{{match}}}'", f"${{{match}}}")
        return node
    elif isinstance(node, dict):
        return {k: remove_quotes(v) for k, v in node.items()}
    else:
        return node


def check_if_assistant_already_exists(
    assistant_name: str, endpoint: str, verify: bool = True
) -> bool:
    """Checks if the assistant already exists in Studio.

    Args:
        assistant_name: The name of the assistant
        endpoint: The studio endpoint
        verify: Whether to verify SSL

    Returns:
        bool: The upload confirmation
    """
    graphql_req = build_get_assistant_by_name_request(assistant_name)

    structlogger.info(
        "rasa.studio.upload.assistant_already_exists",
        event_info="Checking if assistant already exists...",
        assistant_name=assistant_name,
    )

    token = KeycloakTokenReader().get_token()
    res = requests.post(
        endpoint,
        json=graphql_req,
        headers={
            "Authorization": f"{token.token_type} {token.access_token}",
            "Content-Type": "application/json",
        },
        verify=verify,
    )
    response = res.json()["data"]["assistantByName"] or {}
    if results_logger.response_has_id(response):
        structlogger.info(
            "rasa.studio.upload.assistant_already_exists",
            event_info="Assistant already exists.",
        )
        return True

    structlogger.info(
        "rasa.studio.upload.assistant_already_exists", event_info="Assistant not found."
    )
    return False


def build_get_assistant_by_name_request(
    assistant_name: str,
) -> Dict:
    graphql_req = {
        "query": (
            "query AssistantByName($input: AssistantByNameInput!) {"
            " assistantByName(input: $input) {"
            " ... on Assistant { id name mode }"
            " ... on AssistantByName_AssistantNotFound { _ }"
            " }"
            "}"
        ),
        "variables": {
            "input": {
                "assistantName": assistant_name,
            }
        },
    }
    return graphql_req


def _handle_existing_assistant(
    assistant_name: str,
    endpoint: str,
    verify: bool,
    args: argparse.Namespace,
) -> bool:
    """Deal with the case that an assistant with the same name already exists.

    Args:
        assistant_name: The name of the assistant
        endpoint: The studio endpoint
        verify: Whether to verify SSL
        args: The command line arguments

    Returns:
        bool: True if the assistant does not exist and can be created,
              False if the assistant already exists and was linked.
    """
    from rasa.studio.link import handle_link

    if not check_if_assistant_already_exists(assistant_name, endpoint, verify):
        return True

    should_link = questionary.confirm(
        f"An assistant named {assistant_name} already exists in Studio. "
        f"Would you like to link your local project to this existing assistant?"
    ).ask()

    if not should_link:
        rasa.shared.utils.cli.print_error_and_exit("Upload cancelled.")
        return False

    args.assistant_name = assistant_name
    handle_link(args)
    return False



import asyncio
import importlib
import json
from contextlib import asynccontextmanager
from typing import Any, Dict, List, Optional

import openai
import structlog
from jinja2 import Template
from typing_extensions import AsyncGenerator

from rasa.builder import config
from rasa.builder.copilot.constants import (
    COPILOT_PROMPTS_DIR,
    COPILOT_PROMPTS_FILE,
    ROLE_COPILOT,
    ROLE_COPILOT_INTERNAL,
    ROLE_SYSTEM,
    ROLE_USER,
)
from rasa.builder.copilot.exceptions import CopilotStreamError
from rasa.builder.copilot.models import (
    CopilotContext,
    ResponseCategory,
    UsageStatistics,
)
from rasa.builder.document_retrieval.inkeep_document_retrieval import (
    InKeepDocumentRetrieval,
)
from rasa.builder.document_retrieval.models import Document
from rasa.builder.exceptions import (
    DocumentRetrievalError,
)
from rasa.builder.shared.tracker_context import TrackerContext
from rasa.shared.constants import PACKAGE_NAME

structlogger = structlog.get_logger()


class Copilot:
    def __init__(self) -> None:
        self._inkeep_document_retrieval = InKeepDocumentRetrieval()

        # Load both prompt templates up-front
        self._system_message_prompt_template = Template(
            importlib.resources.read_text(
                f"{PACKAGE_NAME}.{COPILOT_PROMPTS_DIR}",
                COPILOT_PROMPTS_FILE,
            )
        )

        # The final stream chunk includes usage statistics.
        self.usage_statistics = UsageStatistics()

    @asynccontextmanager
    async def _get_client(self) -> AsyncGenerator[openai.AsyncOpenAI, None]:
        """Create a fresh OpenAI client, yield it, and always close it."""
        client = openai.AsyncOpenAI(timeout=config.OPENAI_TIMEOUT)
        try:
            yield client
        except Exception as e:
            structlogger.error("copilot.llm_client_error", error=str(e))
            raise
        finally:
            try:
                await client.close()
            except Exception as exc:
                # Closing should not break request processing, but we log it
                structlogger.warning(
                    "copilot.llm_client_close_error",
                    event_info="Failed to close OpenAI client cleanly.",
                    error=str(exc),
                )

    async def search_rasa_documentation(
        self,
        context: CopilotContext,
    ) -> List[Document]:
        """Search Rasa documentation for relevant information.

        Args:
            context: The context of the copilot.

        Returns:
            A list of Document objects. Empty list is returned if the search fails.
        """
        try:
            query = self._create_documentation_search_query(context)
            return await self._inkeep_document_retrieval.retrieve_documents(query)
        except DocumentRetrievalError as e:
            structlogger.error(
                "copilot.search_rasa_documentation.error",
                event_info=(
                    f"Copilot: Searching Rasa documentation for query '{query}' "
                    f"failed with the following error: {e}. Returning empty list."
                ),
                query=query,
                error=str(e),
            )
            return []

    async def generate_response(
        self,
        context: CopilotContext,
    ) -> tuple[AsyncGenerator[str, None], list[Document], str]:
        """Generate a response from the copilot.

        This method performs document retrieval and response generation as a single
        atomic operation. The returned documents are the supporting evidence used
        to generate the response, ensuring consistency between the response content
        and its sources.

        Args:
            context: The context of the copilot.

        Returns:
            A tuple containing the async response stream, relevant documents used
            as supporting evidence for the generated response, and the prompt used.

        Raises:
            CopilotStreamError: If the stream fails.
            Exception: If an unexpected error occurs.
        """
        relevant_documents = await self.search_rasa_documentation(context)
        system_message = await self._create_system_message(context, relevant_documents)
        chat_history = self._create_chat_history_messages(context)
        messages = [system_message, *chat_history]

        return (
            self._stream_response(messages),
            relevant_documents,
            system_message.get("content", ""),
        )

    async def _stream_response(
        self, messages: List[Dict[str, Any]]
    ) -> AsyncGenerator[str, None]:
        """Stream markdown chunks one by one."""
        self.usage_statistics.reset()

        try:
            async with self._get_client() as client:
                stream = await client.chat.completions.create(
                    model=config.OPENAI_MODEL,
                    messages=messages,  # type: ignore
                    temperature=config.OPENAI_TEMPERATURE,
                    stream=True,
                    stream_options={"include_usage": True},
                )
                async for chunk in stream:
                    # The final chunk, which contains the usage statistics,
                    # arrives with an empty `choices` list.
                    if not chunk.choices:
                        self.usage_statistics.update_from_stream_chunk(chunk)
                        # Nothing to yield – continue to the next chunk.
                        continue

                    delta = chunk.choices[0].delta
                    if delta and delta.content:
                        yield delta.content
        except openai.OpenAIError as e:
            structlogger.exception("copilot.stream_response.api_error", error=str(e))
            raise CopilotStreamError(
                "Failed to stream response from OpenAI API."
            ) from e
        except asyncio.TimeoutError as e:
            structlogger.exception(
                "copilot.stream_response.timeout_error", error=str(e)
            )
            raise CopilotStreamError("Request to OpenAI API timed out.") from e
        except Exception as e:
            structlogger.exception(
                "copilot.stream_response.unexpected_error", error=str(e)
            )
            raise

    async def _create_system_message(
        self,
        context: CopilotContext,
        relevant_documents: List[Document],
    ) -> Dict[str, Any]:
        """Render the correct Jinja template based on desired output_type."""
        # Format relevant documentation
        documents = [doc.model_dump() for doc in relevant_documents]

        # Format conversation history
        conversation = self._format_conversation_history(context.tracker_context)

        # Format current state
        current_state = self._format_current_state(context.tracker_context)

        # Render template
        rendered_prompt = self._system_message_prompt_template.render(
            current_conversation=conversation,
            current_state=current_state,
            assistant_logs=context.assistant_logs,
            assistant_files=context.assistant_files,
            documentation_results=documents,
        )
        return {"role": ROLE_SYSTEM, "content": rendered_prompt}

    def _create_chat_history_messages(
        self,
        context: CopilotContext,
    ) -> List[Dict[str, Any]]:
        """Create the chat history messages for the copilot.

        Filter out messages with response_category of GUARDRAILS_POLICY_VIOLATION.
        This will filter out all the user messages that were flagged by guardrails, but
        also the copilot messages that were produced by guardrails.
        """
        # Filter out messages with response_category of GUARDRAILS_POLICY_VIOLATION.
        # This will filter out all the user messages flagged by guardrails, but also the
        # copilot messages that were produced.
        return [
            message.to_openai_format()
            for message in context.copilot_chat_history
            if message.response_category != ResponseCategory.GUARDRAILS_POLICY_VIOLATION
        ]

    @staticmethod
    def _create_documentation_search_query(context: CopilotContext) -> str:
        """Format chat messages between user and copilot for documentation search."""
        result = ""
        role_to_prefix = {
            ROLE_USER: "User",
            ROLE_COPILOT: "Assistant",
            ROLE_COPILOT_INTERNAL: "Copilot Internal Request",
        }
        for message in context.copilot_chat_history:
            prefix = role_to_prefix[message.role]
            text = message.get_text_content().strip()
            if text:
                result += f"{prefix}: {text}\n"
            log_content = message.get_log_content().strip()
            if log_content:
                result += f"{prefix}: {log_content}\n"

        return result

    @staticmethod
    def _format_documents(results: List[Document]) -> Optional[str]:
        """Format documentation search results as JSON dump to be used in the prompt."""
        if not results:
            return None

        formatted_results = {
            "sources": [
                {
                    # Start the reference from 1, not 0.
                    "idx": idx + 1,
                    "title": result.title,
                    "url": result.url,
                    "content": result.content,
                }
                for idx, result in enumerate(results)
            ]
        }
        return json.dumps(formatted_results, ensure_ascii=False, indent=2)

    @staticmethod
    def _format_conversation_history(tracker_context: Optional[TrackerContext]) -> str:
        """Format conversation history from TrackerContext using nested turn structure.

        Args:
            tracker_context: The TrackerContext containing conversation data.

        Returns:
            A JSON string with turns containing user_input, assistant_response,
            and context.

        Example:
            ```json
            {
                "conversation_history": [
                    {
                        "turn_id": 1,
                        "USER": {
                            "text": "I want to transfer money",
                            "predicted_commands": ["start flow", "set slot", ...]
                        },
                        "BOT": [
                            {"text": "How much would you like to transfer?"}
                        ],
                        "other_tracker_events": [
                            {
                                "event": "action_executed",
                                "data": {"action_name": "action_ask_amount"}
                            },
                            {
                                "event": "slot_set",
                                "data": {
                                    "slot_name": "amount_of_money",
                                    "slot_value": 100,
                                },
                            }
                        ]
                    }
                ]
            }
            ```
        """
        conversation_history: Dict[str, Any] = {
            "conversation_history": [],
        }

        if not tracker_context or not tracker_context.conversation_turns:
            return json.dumps(conversation_history, ensure_ascii=False, indent=2)

        conversation_turns: List[Dict[str, Any]] = []
        user_prefix = "USER"
        assistant_prefix = "BOT"

        for turn_idx, turn in enumerate(tracker_context.conversation_turns, 1):
            turn_data: Dict[str, Any] = {"turn_id": turn_idx}

            # Add user if present
            if turn.user_message:
                turn_data[user_prefix] = {
                    "text": turn.user_message.text,
                    "predicted_commands": turn.user_message.predicted_commands,
                }

            # Add assistant messages if present
            if turn.assistant_messages:
                turn_data[assistant_prefix] = [
                    {"text": assistant_message.text}
                    for assistant_message in turn.assistant_messages
                ]

            # Add other tracker events
            if turn.context_events:
                other_events = [event.model_dump() for event in turn.context_events]
                turn_data["other_tracker_events"] = other_events

            conversation_turns.append(turn_data)

        conversation_history["conversation_history"] = conversation_turns
        return json.dumps(conversation_history, ensure_ascii=False, indent=2)

    @staticmethod
    def _format_current_state(tracker_context: Optional[TrackerContext]) -> str:
        """Format current state from TrackerContext for LLM consumption.

        Args:
            tracker_context: The TrackerContext containing current state data.

        Returns:
            A JSON string containing the current state information.
        """
        if not tracker_context or not tracker_context.current_state:
            return json.dumps({}, ensure_ascii=False, indent=2)
        current_state = tracker_context.current_state.model_dump()
        return json.dumps(current_state, ensure_ascii=False, indent=2)



from typing import Any, Dict, List, Optional

from rasa.core.actions.action import Action
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.dialogue_understanding.stack.frames import ChitChatStackFrame
from rasa.shared.core.constants import ACTION_TRIGGER_CHITCHAT
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.trackers import DialogueStateTracker


class ActionTriggerChitchat(Action):
    """Action which triggers a chitchat answer."""

    def name(self) -> str:
        """Return the name of the action."""
        return ACTION_TRIGGER_CHITCHAT

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Run the predicate checks."""
        stack = tracker.stack
        stack.push(ChitChatStackFrame())
        return tracker.create_stack_updated_events(stack)



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.shared.constants import RASA_PATTERN_CANNOT_HANDLE_DEFAULT
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

DATA_KEY_CANNOT_HANDLE_REASON = "cannot_handle_reason"


@dataclass
class CannotHandleCommand(Command):
    """A command to indicate that the bot can't handle the user's input."""

    reason: Optional[Text] = RASA_PATTERN_CANNOT_HANDLE_DEFAULT
    """Reason for cannot handle used in switch-case of the
    cannot handle pattern flow."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "cannot handle"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> CannotHandleCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return CannotHandleCommand(
            data.get("reason", RASA_PATTERN_CANNOT_HANDLE_DEFAULT)
        )

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        if self.reason is not None:
            stack.push(CannotHandlePatternFlowStackFrame(reason=self.reason))
        else:
            stack.push(CannotHandlePatternFlowStackFrame())
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.reason)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CannotHandleCommand):
            return False

        return other.reason == self.reason

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: "CannotHandle()",
            CommandSyntaxVersion.v2: "cannot handle",
            CommandSyntaxVersion.v3: "cannot handle",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> CannotHandleCommand:
        """Converts a DSL string to a command."""
        reason = kwargs.get("data", {}).get(
            DATA_KEY_CANNOT_HANDLE_REASON, RASA_PATTERN_CANNOT_HANDLE_DEFAULT
        )
        return CannotHandleCommand(reason)

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"CannotHandle\(\)",
            CommandSyntaxVersion.v2: r"""^[\s\W\d]*cannot handle['"`]*$""",
            CommandSyntaxVersion.v3: r"""^[\s\W\d]*cannot handle['"`]*$""",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionAddPayee(Action):
    def name(self) -> Text:
        return "action_add_payee"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")
        payee_name = tracker.get_slot("payee_name")
        account_number = tracker.get_slot("account_number")
        sort_code = tracker.get_slot("sort_code")
        payee_type = tracker.get_slot("payee_type")
        reference = tracker.get_slot("reference") or ""

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Add the payee
        success = db.add_payee(
            int(user["id"]),
            payee_name,
            sort_code,
            account_number,
            payee_type,
            reference,
        )

        if success:
            dispatcher.utter_message(
                text=f"Payee {payee_name} has been added successfully."
            )
            return [SlotSet("payee_added", True)]
        else:
            dispatcher.utter_message(text="Failed to add payee. Please try again.")
            return [SlotSet("payee_added", False)]



"""Pydantic models for request/response validation."""

import json
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import structlog
from pydantic import BaseModel, Field, validator

from rasa.cli.scaffold import ProjectTemplateName
from rasa.shared.importers.importer import TrainingDataImporter

structlogger = structlog.get_logger()


class PromptRequest(BaseModel):
    """Request model for prompt-to-bot endpoint."""

    prompt: str = Field(
        ..., min_length=1, max_length=10000, description="The skill description prompt"
    )

    @validator("prompt")
    def validate_prompt(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("Prompt cannot be empty or whitespace only")
        return v.strip()


class TemplateRequest(BaseModel):
    """Request model for template-to-bot endpoint."""

    template_name: ProjectTemplateName = Field(
        ...,
        description=(
            f"The template name to use ({ProjectTemplateName.supported_values()})"
        ),
    )

    @validator("template_name")
    def validate_template_name(cls, v: Any) -> Any:
        if v not in ProjectTemplateName:
            raise ValueError(
                f"Template name must be one of {ProjectTemplateName.supported_values()}"
            )
        return v


class BotDataUpdateRequest(BaseModel):
    """Request model for bot data updates."""

    domain_yml: Optional[str] = Field(None, alias="domain.yml")
    flows_yml: Optional[str] = Field(None, alias="flows.yml")
    config_yml: Optional[str] = Field(None, alias="config.yml")

    class Config:
        """Config for BotDataUpdateRequest."""

        allow_population_by_field_name = True


class BotData(BaseModel):
    """Data of a running assistant."""

    domain: Dict[str, Any] = Field(..., description="The domain of the assistant.")
    flows: Dict[str, Any] = Field(..., description="The flows of the assistant.")


class AssistantInfo(BaseModel):
    """Basic information about the loaded assistant."""

    assistant_id: Optional[str] = Field(
        None,
        description=(
            "Assistant identifier coming from `assistant_id` in the model "
            "metadata (config.yml)."
        ),
    )


class ApiErrorResponse(BaseModel):
    """API error response model."""

    status: Literal["error"] = "error"
    error: str = Field(...)
    details: Optional[Dict[str, Any]] = Field(None)


class ServerSentEventType(str, Enum):
    progress = "progress"
    error = "error"
    _EOF = "_EOF"


class ServerSentEvent(BaseModel):
    """Generic Server-Sent Event payload."""

    event: str = Field(..., description="SSE event name / type")
    data: Dict[str, Any] = Field(
        default_factory=dict,
        description="Arbitrary JSON-serializable payload",
    )

    @classmethod
    def build(cls, event: str, data: Any) -> "ServerSentEvent":
        """General-purpose constructor.

        Args:
            event: The event name (e.g. "progress", "error").
            data: Arbitrary key-value pairs to include in the payload.

        Returns:
            A ServerSentEvent instance with the specified event and data.
        """
        return cls(event=event, data=data)

    @classmethod
    def eof(cls) -> "ServerSentEvent":
        """Helper that returns the special end-of-stream marker."""
        return cls(event=ServerSentEventType._EOF.value, data={})

    def format(self) -> str:
        """Return the text representation used by SSE protocols."""
        return (
            f"event: {self.event}\n"
            f"data: {json.dumps(self.data, separators=(', ', ': '))}\n\n"
        )


class JobStatusEvent(ServerSentEvent):
    """Job status event with special handling for progress and error states."""

    @classmethod
    def from_status(
        cls,
        status: str,
        message: Optional[str] = None,
    ) -> "JobStatusEvent":
        """Factory for job-status events.

        Args:
            status: The job status (e.g. "training", "train_success").
            message: Optional error message for error events.

        Returns:
            A JobStatusEvent instance with the appropriate event type and data.
        """
        event_type = (
            ServerSentEventType.error if message else ServerSentEventType.progress
        )

        payload: Dict[str, Any] = {"status": status}
        if message:
            payload["message"] = message

        return cls(event=event_type.value, data=payload)


class ValidationResult(BaseModel):
    """Result of validation operation."""

    is_valid: bool = Field(...)
    errors: Optional[List[str]] = Field(None)
    warnings: Optional[List[str]] = Field(None)


class TrainingResult(BaseModel):
    """Result of training operation."""

    success: bool = Field(...)
    model_path: Optional[str] = Field(None)
    error: Optional[str] = Field(None)


BotFiles = Dict[str, Optional[str]]


class JobStatus(str, Enum):
    received = "received"
    done = "done"
    error = "error"

    generating = "generating"
    generation_success = "generation_success"
    generation_error = "generation_error"

    training = "training"
    train_success = "train_success"
    train_error = "train_error"

    validating = "validating"
    validation_success = "validation_success"
    validation_error = "validation_error"


class JobCreateResponse(BaseModel):
    job_id: str = Field(...)
    status: JobStatus = JobStatus.received


class TrainingInput(BaseModel):
    """Input for training a model."""

    model_config = {"arbitrary_types_allowed": True}

    importer: TrainingDataImporter = Field(..., description="Training data importer")
    endpoints_file: Path = Field(..., description="Path to the endpoints file")


class AgentStatus(str, Enum):
    """Status of the agent."""

    not_loaded = "not_loaded"
    ready = "ready"
    not_ready = "not_ready"



from datetime import datetime
from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher


class ActionValidatePaymentDate(Action):
    def name(self) -> Text:
        return "action_validate_payment_date"

    def is_future_date(self, payment_date_str: str, current_date_str: str) -> bool:
        current_date = datetime.strptime(current_date_str, "%d/%m/%Y").date()
        payment_date = datetime.strptime(payment_date_str, "%d/%m/%Y").date()
        return payment_date > current_date

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        current_date = tracker.get_slot("current_date")
        payment_date = tracker.get_slot("payment_date")

        if self.is_future_date(payment_date, current_date):
            return [
                SlotSet("valid_payment_date", True),
                SlotSet("future_payment_date", True),
            ]
        else:
            return [
                SlotSet("valid_payment_date", True),
                SlotSet("future_payment_date", False),
            ]



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionCheckSufficientFunds(Action):
    def name(self) -> Text:
        return "action_check_sufficient_funds"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")
        account_from = tracker.get_slot("account_from")
        amount = float(tracker.get_slot("amount"))

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Check if user has sufficient funds
        has_sufficient_funds = db.check_sufficient_funds(
            int(user["id"]), account_from, amount
        )

        if has_sufficient_funds:
            return [SlotSet("sufficient_funds", True)]
        else:
            return [SlotSet("sufficient_funds", False)]



from rasa.shared.constants import (
    MAX_COMPLETION_TOKENS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
    TEMPERATURE_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.utils.llm import (
    DEFAULT_OPENAI_CHAT_MODEL_NAME_ADVANCED,
    DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
)

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_CHAT_MODEL_NAME_ADVANCED,
    TEMPERATURE_CONFIG_KEY: 0.0,
    MAX_COMPLETION_TOKENS_CONFIG_KEY: DEFAULT_OPENAI_MAX_GENERATED_TOKENS,
    TIMEOUT_CONFIG_KEY: 7,
}

LLM_CONFIG_KEY = "llm"
USER_INPUT_CONFIG_KEY = "user_input"

FLOW_RETRIEVAL_KEY = "flow_retrieval"
FLOW_RETRIEVAL_ACTIVE_KEY = "active"
FLOW_RETRIEVAL_EMBEDDINGS_CONFIG_KEY = "embeddings"

FLOW_RETRIEVAL_FLOW_THRESHOLD = 20

COMMAND_PROMPT_FILE_NAME = "command_prompt.jinja2"
LLM_BASED_COMMAND_GENERATOR_CONFIG_FILE = "config.json"

MODEL_NAME_GPT_4O_2024_11_20 = "gpt-4o-2024-11-20"
MODEL_NAME_CLAUDE_3_5_SONNET_20240620 = "claude-3-5-sonnet-20240620"

# JINJA template filters

TO_JSON_ESCAPED_STRING_JINJA_FILTER = "to_json_escaped_string"



from typing import Any, Optional

import structlog
from sanic import Sanic

from rasa.builder import config
from rasa.builder.exceptions import (
    LLMGenerationError,
    ProjectGenerationError,
    TrainingError,
    ValidationError,
)
from rasa.builder.job_manager import JobInfo, job_manager
from rasa.builder.models import (
    JobStatus,
    JobStatusEvent,
)
from rasa.builder.project_generator import ProjectGenerator
from rasa.builder.training_service import (
    train_and_load_agent,
    try_load_existing_agent,
    update_agent,
)
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName

structlogger = structlog.get_logger()


async def push_job_status_event(
    job: JobInfo, status: JobStatus, message: Optional[str] = None
) -> None:
    event = JobStatusEvent.from_status(status=status.value, message=message)
    job.status = status.value
    await job.put(event)


async def run_prompt_to_bot_job(
    app: Any,
    job: JobInfo,
    prompt: str,
) -> None:
    """Run the prompt-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        prompt: The natural language prompt for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1. Generating
        await push_job_status_event(job, JobStatus.generating)
        bot_files = await project_generator.generate_project_with_retries(
            prompt,
            template=ProjectTemplateName.BASIC,
        )
        await push_job_status_event(job, JobStatus.generation_success)

        # 2. Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_agent(project_generator.get_training_input())
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        structlogger.info(
            "bot_builder_service.prompt_to_bot.success",
            files_generated=list(bot_files.keys()),
        )
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "prompt_to_bot_job.training_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "prompt_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )

        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except (ProjectGenerationError, LLMGenerationError) as exc:
        structlogger.debug(
            "prompt_to_bot_job.generation_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "prompt_to_bot_job.unexpected_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_template_to_bot_job(
    app: "Sanic",
    job: JobInfo,
    template_name: ProjectTemplateName,
) -> None:
    """Run the template-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The name of the template to use for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1) Generating
        await push_job_status_event(job, JobStatus.generating)
        await project_generator.init_from_template(template_name)
        bot_files = project_generator.get_bot_files()
        await push_job_status_event(job, JobStatus.generation_success)

        # 2) Training
        await push_job_status_event(job, JobStatus.training)
        agent = await try_load_existing_agent(project_generator.project_folder)
        if agent is None:
            agent = await train_and_load_agent(project_generator.get_training_input())
        else:
            structlogger.info(
                "bot_builder_service.template_to_bot.agent_loaded_from_cache",
            )
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # 3) Done
        structlogger.info(
            "bot_builder_service.template_to_bot.success",
            files_generated=list(bot_files.keys()),
        )
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "template_to_bot_job.training_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "template_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except ProjectGenerationError as exc:
        structlogger.debug(
            "template_to_bot_job.generation_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "template_to_bot_job.unexpected_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_update_files_job(
    app: "Sanic",
    job: JobInfo,
    bot_files: dict,
) -> None:
    project_generator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    try:
        project_generator.update_bot_files(bot_files)

        # 1. Validating
        await push_job_status_event(job, JobStatus.validating)
        training_input = project_generator.get_training_input()
        validation_error = await validate_project(training_input.importer)
        if validation_error:
            raise ValidationError(validation_error)
        await push_job_status_event(job, JobStatus.validation_success)

        # 2. Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_agent(training_input)
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except ValidationError as exc:
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")
        structlogger.debug(
            "update_files_job.validation_error",
            job_id=job.id,
            error=str(exc),
            validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except TrainingError as exc:
        structlogger.debug(
            "update_files_job.train_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback for anything truly unexpected
        structlogger.exception(
            "update_files_job.unexpected_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))



import logging
from abc import abstractmethod
from typing import Any, Dict, List

import litellm
import structlog
from litellm import aembedding, embedding, validate_environment

from rasa.shared.constants import API_BASE_CONFIG_KEY, API_KEY
from rasa.shared.exceptions import (
    ProviderClientAPIException,
    ProviderClientValidationError,
)
from rasa.shared.providers._ssl_verification_utils import (
    ensure_ssl_certificates_for_litellm_non_openai_based_clients,
    ensure_ssl_certificates_for_litellm_openai_based_clients,
)
from rasa.shared.providers.embedding.embedding_response import (
    EmbeddingResponse,
    EmbeddingUsage,
)
from rasa.shared.utils.io import resolve_environment_variables, suppress_logs

structlogger = structlog.get_logger()

_VALIDATE_ENVIRONMENT_MISSING_KEYS_KEY = "missing_keys"


class _BaseLiteLLMEmbeddingClient:
    """An abstract base class for LiteLLM embedding clients.

    This class defines the interface and common functionality for all clients
    based on LiteLLM.

    The class is made private to prevent it from being part of the
    public-facing interface, as it serves as an internal base class
    for specific implementations of clients that are currently based on
    LiteLLM.

    By keeping it private, we ensure that only the derived, concrete
    implementations are exposed to users, maintaining a cleaner and
    more controlled API surface.
    """

    def __init__(self):  # type: ignore
        self._ensure_certificates()

    @property
    @abstractmethod
    def config(self) -> dict:
        """Returns the configuration for that the embedding client in dict form."""
        pass

    @property
    @abstractmethod
    def _litellm_model_name(self) -> str:
        """Returns the model name in LiteLLM format based on the Provider/API type."""
        pass

    @property
    @abstractmethod
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        """Returns a dictionary of extra parameters which include model
        parameters as well as LiteLLM specific input parameters.
        By default, this returns an empty dictionary (no extra parameters).
        """
        return {}

    @property
    def _embedding_fn_args(self) -> Dict[str, Any]:
        """Returns the arguments to be passed to the embedding function."""
        return {
            # Parameters set through config, can override drop_params
            **self._litellm_extra_parameters,
            # Model name is constructed in the LiteLLM format from the provided config
            # Non-overridable to ensure consistency
            "model": self._litellm_model_name,
        }

    def validate_client_setup(self) -> None:
        """Perform client validation. By default only environment variables
        are validated. Override this method to add more validation steps.

        Raises:
            ProviderClientValidationError if validation fails.
        """
        self._validate_environment_variables()

    def _validate_environment_variables(self) -> None:
        """Validate that the required environment variables are set."""
        validation_info = validate_environment(
            self._litellm_model_name,
            api_key=self._litellm_extra_parameters.get(API_KEY),
            api_base=self._litellm_extra_parameters.get(API_BASE_CONFIG_KEY),
        )
        if missing_environment_variables := validation_info.get(
            _VALIDATE_ENVIRONMENT_MISSING_KEYS_KEY
        ):
            event_info = (
                f"Environment variables: {missing_environment_variables} "
                f"not set. Required for API calls."
            )
            structlogger.error(
                "base_litellm_embedding_client.validate_environment_variables",
                event_info=event_info,
                missing_environment_variables=missing_environment_variables,
            )
            raise ProviderClientValidationError(event_info)

    def validate_documents(self, documents: List[str]) -> None:
        """Validates a list of documents to ensure they are suitable for embedding.

        Args:
            documents: List of documents to be validated.

        Raises:
            ValueError: If any document is invalid.
        """
        for doc in documents:
            if not isinstance(doc, str):
                raise ValueError("All documents must be strings.")
            if not doc.strip():
                raise ValueError("Documents cannot be empty or whitespace.")

    @suppress_logs(log_level=logging.WARNING)
    def embed(self, documents: List[str]) -> EmbeddingResponse:
        """Embeds a list of documents synchronously.

        Args:
            documents: List of documents to be embedded.

        Returns:
            List of embedding vectors.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        self.validate_documents(documents)
        try:
            arguments = resolve_environment_variables(self._embedding_fn_args)
            response = embedding(input=documents, **arguments)
            return self._format_response(response)
        except Exception as e:
            raise ProviderClientAPIException(
                message="Failed to embed documents", original_exception=e
            )

    @suppress_logs(log_level=logging.WARNING)
    async def aembed(self, documents: List[str]) -> EmbeddingResponse:
        """Embeds a list of documents asynchronously.

        Args:
            documents: List of documents to be embedded.

        Returns:
            List of embedding vectors.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        self.validate_documents(documents)
        try:
            arguments = resolve_environment_variables(self._embedding_fn_args)
            response = await aembedding(input=documents, **arguments)
            return self._format_response(response)
        except Exception as e:
            raise ProviderClientAPIException(
                message="Failed to embed documents", original_exception=e
            )

    def _format_response(
        self, response: litellm.EmbeddingResponse
    ) -> EmbeddingResponse:
        """Parses the LiteLLM EmbeddingResponse to Rasa format.

        Raises:
            ValueError: If any response data is None.
        """
        # If data is not available (None), raise a ValueError
        if response.data is None:
            message = (
                "Failed to embed documents. Received 'None' " "instead of embeddings."
            )
            structlogger.error(
                "base_litellm_client.format_response.data_is_none",
                message=message,
                response=response.to_dict(),
            )
            raise ValueError(message)

        # Sort the embeddings by the "index" key
        response.data.sort(key=lambda x: x["index"])
        # Extract the embedding vectors
        embeddings = [data["embedding"] for data in response.data]
        formatted_response = EmbeddingResponse(
            data=embeddings,
            model=response.model,
        )

        # Process additional usage information if available
        if response.usage:
            completion_tokens = (
                response.usage.completion_tokens
                if hasattr(response.usage, "completion_tokens")
                else 0
            )
            prompt_tokens = (
                response.usage.prompt_tokens
                if hasattr(response.usage, "prompt_tokens")
                else 0
            )
            total_tokens = (
                response.usage.total_tokens
                if hasattr(response.usage, "total_tokens")
                else 0
            )

            formatted_response.usage = EmbeddingUsage(
                completion_tokens=completion_tokens,
                prompt_tokens=prompt_tokens,
                total_tokens=total_tokens,
            )

        # Log the response with masked data for brevity
        log_response = formatted_response.to_dict()
        log_response["data"] = "Embedding response data not shown here for brevity."
        structlogger.debug(
            "base_litellm_client.formatted_response",
            formatted_response=log_response,
        )
        return formatted_response

    @staticmethod
    def _ensure_certificates() -> None:
        """Configures SSL certificates for LiteLLM. This method is invoked during
        client initialization.

        LiteLLM may utilize `openai` clients or other providers that require
        SSL verification settings through the `SSL_VERIFY` / `SSL_CERTIFICATE`
        environment variables or the `litellm.ssl_verify` /
        `litellm.ssl_certificate` global settings.

        This method ensures proper SSL configuration for both cases.
        """
        ensure_ssl_certificates_for_litellm_non_openai_based_clients()
        ensure_ssl_certificates_for_litellm_openai_based_clients()



import structlog

structlogger = structlog.get_logger()


def get_provider_prefixed_model_name(provider: str, model: str) -> str:
    """
    Returns the model name with the provider prefixed.

    Args:
        provider: The provider of the model.
        model: The model name.

    Returns:
        The model name with the provider prefixed.
    """
    if model and f"{provider}/" not in model:
        return f"{provider}/{model}"
    return model



from typing import Any, Dict, List, Type

from pydantic.main import BaseModel
from sanic_openapi import openapi
from sanic_openapi.openapi3.types import Schema

_SUPPORTED_ATTRIBUTES = frozenset(["format", "enum", "required", "example"])


def _to_schema(
    definition_stack: List[str], schema_def: Dict[str, Any], definitions: Dict[str, Any]
) -> Schema:
    type = schema_def.get("type")

    if type == "object":
        properties_spec = schema_def.get("properties", {})
        properties = {}
        for key in properties_spec:
            properties[key] = _to_schema(
                definition_stack=definition_stack,
                schema_def=properties_spec[key],
                definitions=definitions,
            )
        schema = openapi.Object(
            title=schema_def.get("title"),
            description=schema_def.get("description"),
            required=schema_def.get("required"),
            properties=properties,
        )
    elif type == "array":
        schema = openapi.Array(
            description=schema_def.get("description"),
            required=schema_def.get("required"),
            items=_to_schema(
                definition_stack=definition_stack,
                schema_def=schema_def.get("items"),
                definitions=definitions,
            ),
        )
    elif type is None:
        if allof_spec := schema_def.get("allOf"):  # Model, Enum
            definition = allof_spec[0]["$ref"].split("/")[-1]
            definition_data = definitions.get(definition)
            if definition_data is None:
                schema = openapi.Object(
                    title=definition, description=schema_def.get("description")
                )
            else:
                schema = (
                    _to_schema(
                        definition_stack=definition_stack + [definition],
                        schema_def={**definition_data},
                        definitions=definitions,
                    )
                    if definition not in definition_stack
                    else openapi.Object(
                        title=definition, description=schema_def.get("description")
                    )
                )

        elif anyof_spec := schema_def.get("anyOf"):  # Union
            anyof = []
            for any in anyof_spec:
                if any.get("type"):
                    schema_type_obj = Schema(
                        **{
                            "type": any.get("type"),
                            "description": any.get("description"),
                        }
                    )
                    anyof.append(schema_type_obj)
                else:
                    definition = any["$ref"].split("/")[-1]
                    if definition not in definition_stack:
                        definition_data = definitions.get(definition)
                        if definition_data is not None:
                            anyof.append(
                                _to_schema(
                                    definition_stack=definition_stack + [definition],
                                    schema_def=definition_data,
                                    definitions=definitions,
                                )
                            )
                        else:
                            anyof.append(
                                openapi.Object(
                                    title=definition,
                                    description=schema_def.get(
                                        "description", definition
                                    ),
                                    properties={},
                                )
                            )
                    else:
                        anyof.append(
                            openapi.Object(
                                title=definition,
                                description=schema_def.get("description", definition),
                                properties={},
                            )
                        )
            schema = Schema(anyOf=anyof)
        elif ref := schema_def.get("$ref"):  # $ref
            definition = ref.split("/")[-1]
            definition_data = definitions.get(definition)
            if definition_data is not None:
                schema = _to_schema(
                    definition_stack=definition_stack,
                    schema_def=definition_data,
                    definitions=definitions,
                )
            else:
                schema = openapi.Object(
                    title=definition, description=schema_def.get("description")
                )
        else:  # Any type
            schema = Schema(
                **{"type": "object", "description": schema_def.get("description")}
            )

    else:
        schema_spec = {
            "type": schema_def.get("type"),
            "description": schema_def.get("description"),
        }
        for spec in _SUPPORTED_ATTRIBUTES:
            if schema_def.get(spec):
                schema_spec[spec] = schema_def.get(spec)
        schema = Schema(**schema_spec)

    return schema


def model_to_schema(model: Type[BaseModel]) -> Schema:
    schema = model.model_json_schema()
    # Handle both $defs (newer JSON Schema) and definitions (older JSON Schema)
    definitions = schema.get("$defs") or schema.get("definitions") or {}
    return _to_schema(
        definition_stack=[],
        schema_def=dict(
            filter(lambda key: key[0] not in ("definitions", "$defs"), schema.items())
        ),
        definitions=definitions,
    )



import asyncio
import json
import os
from contextlib import asynccontextmanager
from typing import AsyncGenerator, List, Optional

import openai
import structlog
from openai.types.chat import ChatCompletion

from rasa.builder import config
from rasa.builder.copilot.constants import ROLE_USER
from rasa.builder.document_retrieval.constants import (
    INKEEP_API_KEY_ENV_VAR,
    INKEEP_DOCUMENT_RETRIEVAL_MODEL,
    INKEEP_RAG_RESPONSE_SCHEMA_PATH,
)
from rasa.builder.document_retrieval.models import Document
from rasa.builder.exceptions import DocumentRetrievalError
from rasa.shared.utils.io import read_json_file

structlogger = structlog.get_logger()


class InKeepDocumentRetrieval:
    """Handles the document retrieval from InKeep AI."""

    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self._rag_schema = read_json_file(INKEEP_RAG_RESPONSE_SCHEMA_PATH)
        self._api_key = api_key or os.getenv(INKEEP_API_KEY_ENV_VAR)

    @property
    def api_key(self) -> str:
        """Resolve the correct API key/token based on proxy usage."""
        using_proxy = bool(config.HELLO_LLM_PROXY_BASE_URL)
        if using_proxy:
            if not config.RASA_PRO_LICENSE:
                raise DocumentRetrievalError(
                    "HELLO_LLM_PROXY_BASE_URL is set but RASA_PRO_LICENSE is missing. "
                    "Proxy requires a Rasa Pro license token for authentication."
                )
            return config.RASA_PRO_LICENSE

        if not self._api_key:
            raise DocumentRetrievalError(
                "INKEEP_API_KEY is missing. Provide it via env INKEEP_API_KEY "
                "or pass api_key= to InKeepDocumentRetrieval."
            )
        return self._api_key

    async def retrieve_documents(
        self, query: str, temperature: float = 0.0, timeout: float = 30.0
    ) -> List[Document]:
        """Retrieve relevant documents using InKeep AI based on the given query.

        Args:
            query: The search query
            temperature: Controls randomness in generation (0.0 for deterministic)
            timeout: Timeout for the API call

        Returns:
            List of Document objects containing retrieved content

        Raises:
            DocumentRetrievalError: When document retrieval fails due to:
                - Empty response from InKeep AI API
                - OpenAI API errors (authentication, rate limiting, etc.)
                - Request timeout
                - Unexpected errors during API communication
        """
        try:
            response = await self._call_inkeep_rag_api(
                query=query,
                temperature=temperature,
                timeout=timeout,
            )
            documents = self._parse_documents_from_response(response)
            return documents
        except DocumentRetrievalError as e:
            structlogger.error(
                "inkeep_document_retrieval.retrieve_documents.error",
                event_info="InKeep Document Retrieval: Error",
                query=query,
                error=str(e),
            )
            raise e

    async def _call_inkeep_rag_api(
        self, query: str, temperature: float, timeout: float
    ) -> ChatCompletion:
        """Call InKeep AI RAG's API endpoint and return the response content.

        Args:
            query: The search query to send to InKeep
            temperature: Controls randomness in generation (0.0 for deterministic)
            timeout: Timeout for the API call

        Returns:
            The response content from InKeep AI. The response is made of the retrieved
            documents.

        Raises:
            LLMGenerationError: If the API call fails or returns invalid response
        """
        try:
            async with self._get_client() as client:
                response = await client.chat.completions.create(
                    model=INKEEP_DOCUMENT_RETRIEVAL_MODEL,
                    messages=[{"role": ROLE_USER, "content": query}],
                    temperature=temperature,
                    timeout=timeout,
                    response_format={
                        "type": "json_schema",
                        "json_schema": self._rag_schema,
                    },
                )

                if not response.choices[0].message.content:
                    structlogger.warning(
                        "inkeep_document_retrieval.empty_response",
                        event_info="InKeep AI returned an empty response. ",
                        query=query,
                        response_content=response.choices[0].message.content,
                    )
                    raise DocumentRetrievalError(
                        "InKeep Document Retrieval: Empty response"
                    )

                return response

        except openai.OpenAIError as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.api_error",
                event_info="InKeep Document Retrieval: API error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(f"InKeep Document Retrieval: API error: {e}")
        except asyncio.TimeoutError as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.timeout_error",
                event_info="InKeep Document Retrieval: Timeout error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(f"InKeep AI request timed out: {e}")
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.call_inkeep_rag_api.error",
                event_info="InKeep Document Retrieval: Error",
                query=query,
                error=e,
            )
            raise DocumentRetrievalError(
                f"InKeep Document Retrieval: Unexpected error: {e}"
            )

    @asynccontextmanager
    async def _get_client(self) -> AsyncGenerator[openai.AsyncOpenAI, None]:
        """Get or create client that handles the API calls to InKeep AI."""
        # Ensure trailing slash to match client expectations/tests
        base_url = f"{config.INKEEP_BASE_URL.rstrip('/')}/"
        client = openai.AsyncOpenAI(base_url=base_url, api_key=self.api_key)
        structlogger.debug("inkeep_document_retrieval._get_client", base_url=base_url)

        try:
            yield client
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.client_error",
                event_info="InKeep Document Retrieval: Client error",
                error=str(e),
            )
            raise

    def _parse_documents_from_response(
        self, response: ChatCompletion
    ) -> List[Document]:
        """Parse the InKeep AI response into Document objects.

        Args:
            response: ChatCompletion response from InKeep AI's RAG model.

        Returns:
            List of Document objects. Empty list is returned if the response is empty
            or if the response is invalid.
        """
        try:
            content = response.choices[0].message.content
            if not content:
                return []

            response_data = json.loads(content)
            documents = []

            for item in response_data.get("content", []):
                try:
                    document = Document.from_inkeep_rag_response(item)
                    documents.append(document)
                except Exception as e:
                    structlogger.warning(
                        "inkeep_document_retrieval.invalid_document_skipped",
                        event_info=(
                            "InKeep Document Retrieval: Invalid document structure "
                            "skipped. Returning empty list for this item."
                        ),
                        error=str(e),
                        item=item,
                    )
                    # Continue processing other items, skip this invalid one
                    continue

            return documents

        except json.JSONDecodeError as e:
            structlogger.warning(
                "inkeep_document_retrieval.parse_documents_from_response"
                ".parse_response_failed",
                event_info=(
                    "InKeep Document Retrieval: Parse response failed. "
                    "Returning empty list.",
                ),
                error=str(e),
            )
            return []
        except Exception as e:
            structlogger.error(
                "inkeep_document_retrieval.parse_documents_from_response.error",
                event_info=(
                    "InKeep Document Retrieval: Parse response error. "
                    "Returning empty list.",
                ),
                error=str(e),
            )
            return []



from __future__ import annotations

from typing import Protocol, runtime_checkable


@runtime_checkable
class ClientConfig(Protocol):
    """
    Protocol for the client config that specifies the interface for interacting
    with the API.
    """

    @classmethod
    def from_dict(cls, config: dict) -> ClientConfig:
        """
        Initializes the client config with the given configuration.

        This class method should be implemented to parse the given
        configuration and create an instance of an client config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            ClientConfig
        """
        ...

    def to_dict(self) -> dict:
        """
        Returns the configuration for that the client config is initialized with.

        This method should be implemented to return a dictionary containing
        the configuration settings for the client config.

        Returns:
            dictionary containing the configuration settings for the client config.
        """
        ...

    @staticmethod
    def resolve_config_aliases(config: dict) -> dict:
        """
        Resolve any potential aliases in the configuration.

        This method should be implemented to resolve any potential aliases in the
        configuration.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            dictionary containing the resolved configuration settings for the
            client config.
        """
        ...



from typing import List

from langchain_core.embeddings.embeddings import Embeddings

from rasa.shared.providers.embedding.embedding_client import EmbeddingClient


class _LangchainEmbeddingClientAdapter(Embeddings):
    """
    Temporary adapter to bridge differences between LiteLLM and LangChain.

    Clients instantiated with `embedder_factory` follow our new EmbeddingClient
    protocol, but `langchain`'s vector stores require an `Embeddings` type
    client. This adapter extracts and returns the necessary part of the output
    from our LiteLLM-based clients.

    This adapter will be removed in ticket:
    https://rasahq.atlassian.net/browse/ENG-1220
    """

    def __init__(self, client: EmbeddingClient):
        self._client = client

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed search docs.

        Args:
            texts: List of text to embed.

        Returns:
            List of embeddings.
        """
        response = self._client.embed(documents=texts)
        embedding_vector = response.data
        return embedding_vector

    def embed_query(self, text: str) -> List[float]:
        """Embed query text.

        Args:
            text: Text to embed.

        Returns:
            Embedding.
        """
        response = self._client.embed(documents=[text])
        embedding_vector = response.data[0]
        return embedding_vector

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous Embed search docs.

        Args:
            texts: List of text to embed.

        Returns:
            List of embeddings.
        """
        response = await self._client.aembed(documents=texts)
        embedding_vector = response.data
        return embedding_vector

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text.

        Args:
            text: Text to embed.

        Returns:
            Embedding.
        """
        response = await self._client.aembed(documents=[text])
        embedding_vector = response.data[0]
        return embedding_vector



from typing import Dict

from pydantic import BaseModel


class CommandMetrics(BaseModel):
    tp: int
    fp: int
    fn: int
    total_count: int

    @staticmethod
    def _safe_divide(numerator: float, denominator: float) -> float:
        """Safely perform division, returning 0.0 if the denominator is zero."""
        return numerator / denominator if denominator > 0 else 0.0

    def get_precision(self) -> float:
        return self._safe_divide(self.tp, self.tp + self.fp)

    def get_recall(self) -> float:
        return self._safe_divide(self.tp, self.tp + self.fn)

    def get_f1_score(self) -> float:
        precision = self.get_precision()
        recall = self.get_recall()

        return self._safe_divide(2 * precision * recall, precision + recall)

    def as_dict(self) -> Dict[str, float]:
        return {
            "tp": self.tp,
            "fp": self.fp,
            "fn": self.fn,
            "precision": self.get_precision(),
            "recall": self.get_recall(),
            "f1_score": self.get_f1_score(),
            "total_count": self.total_count,
        }



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionCheckBalance(Action):
    def name(self) -> Text:
        return "action_check_balance"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")
        account_number = tracker.get_slot("account")

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Get account information
        account = db.get_account_by_user_and_number(int(user["id"]), account_number)
        if not account:
            dispatcher.utter_message(text="Account not found.")
            return []

        current_balance = float(account["balance"])

        message = f"The balance on that account is: ${current_balance}"
        dispatcher.utter_message(text=message)
        return []



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionAskAccountFrom(Action):
    def name(self) -> Text:
        return "action_ask_account_from"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Get all accounts for the user
        user_id = int(user["id"])
        accounts = db.get_accounts_by_user(user_id)
        print(f"DEBUG: Accounts found: {accounts}")
        if not accounts:
            dispatcher.utter_message(text="No accounts found for this user.")
            return []

        buttons = [
            {
                "content_type": "text",
                "title": (
                    f"{account['type'].title()} (Balance: "
                    f"${float(account['balance']):.2f})"
                ),
                "payload": str(account["number"]),
            }
            for account in accounts
        ]
        message = "Which account would you like to transfer money from?"
        dispatcher.utter_message(text=message, buttons=buttons)

        return []



from typing import Literal

LAKERA_API_KEY_ENV_VAR = "LAKERA_API_KEY"
LAKERA_GUARD_ENDPOINT = "guard"
LAKERA_GUARD_RESULTS_ENDPOINT = "guard/results"

BLOCK_SCOPE_USER: Literal["user"] = "user"
BLOCK_SCOPE_PROJECT: Literal["project"] = "project"
BlockScope = Literal["user", "project"]



DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256

LOG_COMPONENT_SOURCE_METHOD_INIT = "init"
LOG_COMPONENT_SOURCE_METHOD_FINGERPRINT_ADDON = "fingerprint_addon"



import asyncio
import re
from csv import Error as CSVError
from dataclasses import dataclass, field
from importlib.resources import read_text
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, List, Optional

import pandas as pd
import ruamel
import structlog
from jinja2 import Template
from ruamel.yaml.scanner import ScannerError

from rasa import telemetry
from rasa.e2e_test.constants import KEY_TEST_CASES
from rasa.e2e_test.e2e_config import LLME2ETestConverterConfig
from rasa.e2e_test.utils.validation import read_e2e_test_schema
from rasa.exceptions import RasaException
from rasa.shared.utils.llm import llm_factory
from rasa.shared.utils.yaml import (
    YamlValidationException,
    validate_yaml_data_using_schema_with_assertions,
)

structlogger = structlog.get_logger()

DEFAULT_E2E_OUTPUT_TESTS_DIRECTORY = "e2e_tests"
E2E_TEST_MODULE = "rasa.e2e_test"
DEFAULT_E2E_TEST_CONVERTER_PROMPT_PATH = "e2e_test_converter_prompt.jinja2"

CSV = ".csv"
XLSX = ".xlsx"
XLS = ".xls"
EXCEL_EXTENSIONS = [XLS, XLSX]
ALLOWED_EXTENSIONS = [CSV, *EXCEL_EXTENSIONS]

NUMBER_OF_LLM_ATTEMPTS = 3


@dataclass
class ConversationEntry:
    """Class for storing an entry in a conversation."""

    data: Dict[str, Any]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ConversationEntry":
        return cls(data=data)

    def as_dict(self) -> Dict[str, Any]:
        return self.data


@dataclass
class Conversation:
    """Class for storing a single conversation."""

    entries: List[ConversationEntry] = field(default_factory=list)

    def add_entry(self, entry: ConversationEntry) -> None:
        self.entries.append(entry)

    def as_list(self) -> List[Dict[str, Any]]:
        return [entry.as_dict() for entry in self.entries]


@dataclass
class Conversations:
    """Class for storing multiple conversations."""

    conversations: List[Conversation] = field(default_factory=list)

    def add_conversation(self, conversation: Conversation) -> None:
        self.conversations.append(conversation)

    def as_list(self) -> List[List[Dict[str, Any]]]:
        return [conversation.as_list() for conversation in self.conversations]


class E2ETestConverter:
    """E2ETestConverter class is responsible for reading input CSV or XLS/XLSX files,
    splitting the data into distinct conversations, and converting them into test cases.
    """

    def __init__(
        self,
        path: str,
        llm_config: LLME2ETestConverterConfig,
        sheet_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the E2ETestConverter with necessary parameters.

        Args:
            path (str): Path to the input file.
            sheet_name (str): Name of the sheet in XLSX file.
            llm_config (LLME2ETestConverterConfig): LLM configuration .
        """
        self.input_path: str = path
        self.sheet_name: Optional[str] = sheet_name
        self.input_file_extension: str = self.get_and_validate_input_file_extension()
        self.e2e_schema = read_e2e_test_schema()
        self.prompt_template: str = read_text(
            E2E_TEST_MODULE, DEFAULT_E2E_TEST_CONVERTER_PROMPT_PATH
        )
        self.llm_config = llm_config
        self.default_llm_config_dict = LLME2ETestConverterConfig.get_default_config()

    @staticmethod
    def remove_markdown_code_syntax(markdown_string: str) -> str:
        """Remove Markdown code formatting from the string.

        Args:
            markdown_string (str): string to be parsed.

        Returns:
            str: Parsed string.

        Example:
            >>> markdown_str = '''yaml\nkey: value\n'''
            >>> E2ETestConverter.remove_markdown_code_syntax(markdown_str)
            'key: value'
        """
        if not markdown_string:
            return ""

        dedented_string = dedent(markdown_string)
        regex = r"^```.*\n|```$"
        return re.sub(regex, "", dedented_string, flags=re.MULTILINE).strip()

    @staticmethod
    def split_data_into_conversations(
        input_data: List[Dict],
    ) -> Conversations:
        """Splits the data into conversations using empty row as a separator.

        Arguments:
            input_data (List[Dict]): The list of rows of the input file.

        Returns:
            Conversations: List of conversations
        """
        conversations = Conversations()
        conversation = Conversation()

        # Iterate through each row until the empty row separator is hit
        for row in input_data:
            # Remove empty values from the row
            row = {key: value for key, value in row.items() if value}

            if row:
                conversation_entry = ConversationEntry.from_dict(row)
                conversation.add_entry(conversation_entry)
            else:
                # If the current conversation exists, add it to the list
                if conversation:
                    conversations.add_conversation(conversation)
                    conversation = Conversation()

        # Add last conversation to the list as it might not have a separator after it
        if conversation:
            conversations.add_conversation(conversation)

        return conversations

    async def generate_llm_response(self, prompt: str) -> Optional[str]:
        """Use LLM to generate a response.

        Args:
            prompt (str): the prompt to send to the LLM.

        Returns:
            Optional[str]: Generated response.
        """
        llm = llm_factory(self.llm_config.as_dict(), self.default_llm_config_dict)

        try:
            llm_response = await llm.acompletion(prompt)
            return llm_response.choices[0]
        except Exception as exc:
            structlogger.debug("e2e_test_generator.llm_response_error", exc=exc)
            return None

    def is_yaml_valid(self, yaml_string: str) -> bool:
        """Validate the string against the YAML format.

        Args:
            yaml_string (str): String to be validated

        Returns:
            bool: True if valid, False otherwise
        """
        try:
            yaml_data = ruamel.yaml.safe_load(yaml_string)
            validate_yaml_data_using_schema_with_assertions(
                yaml_data={KEY_TEST_CASES: yaml_data}, schema_content=self.e2e_schema
            )
            return True
        except (YamlValidationException, ScannerError) as exc:
            structlogger.debug("e2e_test_generator.yaml_string_invalid", exc=exc)
            return False

    def get_and_validate_input_file_extension(self) -> str:
        """Validates the input file extension and checks for required properties
        like sheet name for XLSX files.

        Raises:
            RasaException: If the directory path is provided, file extension
            is not supported or sheet name is missing.
        """
        input_file_extension = Path(self.input_path).suffix.lower()
        if not input_file_extension:
            raise RasaException(
                "The path must point to a specific file, not a directory."
            )

        if input_file_extension not in ALLOWED_EXTENSIONS:
            raise RasaException(
                f"Unsupported file type: {input_file_extension}. "
                "Please use .csv or .xls/.xlsx"
            )

        if input_file_extension in EXCEL_EXTENSIONS and self.sheet_name is None:
            raise RasaException(
                f"Please provide a sheet name for the {input_file_extension} file."
            )

        return input_file_extension

    def read_csv(self) -> List[Dict]:
        """Reads a CSV file and converts it into a list of dictionaries.

        Returns:
            List[Dict]: Parsed data from the CSV file as a list of dictionaries.
        """
        data_frame = pd.read_csv(self.input_path, sep=None, engine="python")
        data_frame = data_frame.fillna("")
        return data_frame.to_dict(orient="records")

    def read_xlsx(self) -> List[Dict]:
        """Reads an XLSX file and converts it into a list of dictionaries.

        Returns:
            List[Dict]: Parsed data from the XLSX file as a list of dictionaries.
        """
        data_frame = pd.read_excel(self.input_path, sheet_name=self.sheet_name)
        data_frame = data_frame.fillna("")
        return data_frame.to_dict(orient="records")

    def read_file(self) -> List[Dict]:
        """Calls the appropriate file reading method based on the file extension.

        Raises:
            RasaException: If the file could not be read.

        Returns:
            List[Dict]: Parsed data from the input file as a list of dictionaries.
        """
        extension_to_method = {
            CSV: self.read_csv,
            XLS: self.read_xlsx,
            XLSX: self.read_xlsx,
        }

        try:
            structlogger.debug(
                "e2e_test_generator.read_file",
                input_file_extension=self.input_file_extension,
            )
            return extension_to_method[self.input_file_extension]()
        except pd.errors.ParserError:
            raise RasaException("The file could not be read due to a parsing error.")
        except pd.errors.EmptyDataError:
            raise RasaException("The file is empty and could not be read.")
        except CSVError:
            raise RasaException("There was an error with reading the CSV file.")
        except ValueError:
            raise RasaException("There was a value error while reading the file.")

    def render_template(self, conversation: Conversation) -> str:
        """Renders a jinja2 template.

        Arguments:
            conversation (Conversation): The list of rows of a conversation.

        Returns:
            str: The rendered template string.
        """
        kwargs = {"conversation": conversation.as_list()}
        return Template(self.prompt_template).render(**kwargs)

    async def convert_single_conversation_into_test(
        self, conversation: Conversation
    ) -> str:
        """Uses LLM to convert a conversation to a YAML test case.

        Arguments:
            conversation (Conversation): The list of rows of a conversation.

        Returns:
            str: YAML representation of the conversation.
        """
        yaml_test_case = ""
        prompt_template = self.render_template(conversation)

        for idx in range(NUMBER_OF_LLM_ATTEMPTS):
            response = await self.generate_llm_response(prompt_template)
            yaml_test_case = self.remove_markdown_code_syntax(response)
            if not self.is_yaml_valid(yaml_test_case):
                if idx < NUMBER_OF_LLM_ATTEMPTS - 1:
                    event_info = "Invalid YAML format. Retrying the request."
                else:
                    event_info = (
                        "Invalid YAML format. Failed to convert the "
                        "conversation into a test case."
                    )
                structlogger.debug(
                    "e2e_test_generator.invalid_yaml_format",
                    event_info=event_info,
                )
                continue
            break

        return yaml_test_case

    async def convert_conversations_into_tests(
        self, conversations: Conversations
    ) -> str:
        """Generates test cases from the parsed data.

        Arguments:
            conversations (Conversations): The list of conversation rows.

        Returns:
            str: YAML representation of the test cases.
        """
        # Convert all conversations into YAML test cases asynchronously.
        tasks = [
            asyncio.ensure_future(
                self.convert_single_conversation_into_test(conversation)
            )
            for conversation in conversations.conversations
        ]
        results = await asyncio.gather(*tasks)
        filtered_results = [result for result in results if result]

        telemetry.track_e2e_test_conversion_completed(
            file_type=self.input_file_extension, test_case_count=len(filtered_results)
        )

        structlogger.debug("e2e_test_generator.test_generation_finished")
        return "\n".join(filtered_results)

    def run(self) -> str:
        """Executes the E2E test conversion process: reads the file,
        splits the data into conversations, and generates test cases.
        """
        input_data = self.read_file()
        conversations = self.split_data_into_conversations(input_data)
        return asyncio.run(self.convert_conversations_into_tests(conversations))



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

import structlog
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from pydantic import BaseModel, Field, field_serializer, model_validator
from typing_extensions import Annotated

from rasa.builder.copilot.constants import (
    ROLE_ASSISTANT,
    ROLE_COPILOT,
    ROLE_COPILOT_INTERNAL,
    ROLE_USER,
)
from rasa.builder.models import ServerSentEvent
from rasa.builder.shared.tracker_context import TrackerContext

structlogger = structlog.get_logger()


class ResponseCompleteness(Enum):
    """Enum for response completeness levels."""

    TOKEN = "token"  # Streaming token/chunk
    COMPLETE = "complete"  # Complete response (e.g., templated responses)


class ResponseCategory(Enum):
    """Enum for different categories of responses."""

    # Copilot generated content
    COPILOT = "copilot"
    REFERENCE = "reference"
    REFERENCE_ENTRY = "reference_entry"
    # When Copilot detects a roleplay request / intent
    ROLEPLAY_DETECTION = "roleplay_detection"
    # When Copilot detects an out-of-scope request
    OUT_OF_SCOPE_DETECTION = "out_of_scope_detection"
    # When Copilot does not understand what caused the error
    ERROR_FALLBACK = "error_fallback"
    # When a policy violation is detected
    GUARDRAILS_POLICY_VIOLATION = "guardrails_policy_violation"
    # When Copilot access is blocked after repeated violations
    GUARDRAILS_BLOCKED = "guardrails_blocked"
    # When Copilot detects request for KB content
    KNOWLEDGE_BASE_ACCESS_REQUESTED = "knowledge_base_access_requested"
    # When Copilot analyzes error logs and provides suggestions
    TRAINING_ERROR_LOG_ANALYSIS = "training_error_log_analysis"
    E2E_TESTING_ERROR_LOG_ANALYSIS = "e2e_testing_error_log_analysis"

    # Conversation history signature
    SIGNATURE = "signature"


class BaseContent(BaseModel):
    type: str


class LinkContent(BaseContent):
    type: Literal["link"]
    url: str
    label: str


class ButtonContent(BaseContent):
    type: Literal["button"]
    payload: str
    label: str


class TextContent(BaseContent):
    type: Literal["text"]
    text: str


class CodeContent(BaseContent):
    type: Literal["code"]
    text: str


class FileContent(BaseContent):
    type: Literal["file"]
    file_path: str
    file_content: str


class LogContent(BaseContent):
    type: Literal["log"]
    content: str = Field(..., description="Logs, error messages, stack traces, etc.")
    context: Optional[str] = Field(
        None,
        description=(
            "Additional, optional context description for the logs "
            "(e.g., 'training session', 'e2e testing run', 'deployment process')"
        ),
    )
    metadata: Dict[str, Any] = Field(
        default_factory=dict,
        description="Additional metadata about the log.",
    )


class EventContent(BaseModel):
    type: Literal["event"]
    event: str = Field(..., description="The event's type_name")

    event_data: Dict[str, Any] = Field(
        default_factory=dict, description="Contains event-specific data fields."
    )

    @model_validator(mode="before")
    @classmethod
    def _collect_event_data(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        generic = {"type", "event"}
        data["event_data"] = {
            key: data.pop(key) for key in list(data.keys()) if key not in generic
        }
        return data

    class Config:
        extra = "forbid"


ContentBlock = Annotated[
    Union[
        TextContent,
        CodeContent,
        FileContent,
        LogContent,
        EventContent,
        LinkContent,
        ButtonContent,
    ],
    Field(
        discriminator="type",
        description=(
            "The content of the message. "
            "The content is expected to be a list of content blocks. "
            "The content blocks are expected to be one of the following types: "
            "text, link, code, or file."
        ),
    ),
]


class CopilotChatMessage(BaseModel):
    """Model for a single chat messages between the user and the copilot."""

    role: str = Field(
        ...,
        pattern=f"^({ROLE_USER}|{ROLE_COPILOT}|{ROLE_COPILOT_INTERNAL})$",
        description="The role of the message sender.",
    )
    content: List[ContentBlock]
    response_category: Optional[ResponseCategory] = Field(
        None,
        description=(
            "The category/source of this message. For user role messages, only `None` "
            "or `GUARDRAILS_POLICY_VIOLATION` are allowed. For copilot role messages, "
            "any category is permitted."
        ),
    )

    @model_validator(mode="after")
    def validate_response_category_for_role(self) -> "CopilotChatMessage":
        """Validate value of response_category for the role of the message.

        For 'user' role messages, only None or GUARDRAILS_POLICY_VIOLATION are allowed.
        For 'copilot' role messages, any category is permitted.
        For 'rasa_internal' role messages, any category is permitted.
        """
        if (
            self.role == ROLE_USER
            and self.response_category is not None
            and self.response_category != ResponseCategory.GUARDRAILS_POLICY_VIOLATION
        ):
            message = (
                f"User role messages can only have response_category of `None` or "
                f"`{ResponseCategory.GUARDRAILS_POLICY_VIOLATION}`, "
                f"got `{self.response_category}`."
            )
            structlogger.error(
                "copilot_chat_message.validate_response_category_for_role"
                ".invalid_response_category",
                event_info=message,
                response_category=self.response_category,
                role=self.role,
            )
            raise ValueError(message)

        return self

    @field_serializer("response_category", when_used="always")
    def _serialize_response_category(
        self, v: Optional[ResponseCategory]
    ) -> Optional[str]:
        """Serializing CopilotChatMessage, response_category should be a string."""
        return None if v is None else v.value

    def get_text_content(self) -> str:
        """Concatenate all 'text' content blocks into a single string."""
        return "\n".join(
            content_block.text
            for content_block in self.content
            if isinstance(content_block, TextContent)
        )

    def get_log_content(self) -> str:
        """Concatenate all 'log' content blocks into a single string."""
        return "\n".join(
            content_block.content
            for content_block in self.content
            if isinstance(content_block, LogContent)
        )

    def to_openai_format(self) -> Dict[str, Any]:
        """Convert to OpenAI message format for API calls."""
        role_to_openai_format = {
            ROLE_USER: self._user_message_to_openai_format,
            ROLE_COPILOT: self._copilot_message_to_openai_format,
            ROLE_COPILOT_INTERNAL: self._copilot_message_to_openai_format,
        }
        return role_to_openai_format[self.role]()

    def _user_message_to_openai_format(self) -> Dict[str, Any]:
        role = self._map_role_to_openai()
        content = self.get_text_content()
        return {"role": role, "content": content}

    def _copilot_message_to_openai_format(self) -> Dict[str, Any]:
        role = self._map_role_to_openai()
        # For now the Copilot responds only with the text content and all the content
        # is formatted as a markdown.
        # TODO: Once we start predicting the files, and expecting other content blocks
        #       we should update this.
        content = self.get_text_content()
        return {"role": role, "content": content}

    def _map_role_to_openai(self) -> str:
        """Map internal roles to OpenAI-compatible roles."""
        role_mapping = {
            ROLE_USER: ROLE_USER,
            ROLE_COPILOT: ROLE_ASSISTANT,
            ROLE_COPILOT_INTERNAL: ROLE_USER,
        }
        if self.role not in role_mapping.keys():
            structlogger.error(
                "copilot_chat_message.to_openai_format.invalid_role",
                event_info=(
                    f"Invalid role: `{self.role}`. "
                    f"Only {', '.join(role_mapping.keys())} roles are supported."
                ),
                role=self.role,
            )
            raise ValueError(f"Invalid role: {self.role}")

        return role_mapping[self.role]


class CopilotRequest(BaseModel):
    """Request model for the copilot endpoint."""

    copilot_chat_history: List[CopilotChatMessage] = Field(
        ...,
        description=(
            "The chat history between the user and the copilot. "
            "Used to generate a new response based on the previous conversation."
        ),
    )
    session_id: str = Field(
        ...,
        description=(
            "The session ID of chat session with the assistant. "
            "Used to fetch the conversation from the tracker."
        ),
    )
    history_signature: Optional[str] = Field(
        default=None,
        description="HMAC signature (base64url) for the provided chat history.",
    )
    signature_version: Optional[str] = Field(
        default=None,
        description='Signature scheme version (e.g. "v1").',
    )

    @property
    def last_message(self) -> Optional[CopilotChatMessage]:
        """Get the last message from the copilot chat history."""
        if not self.copilot_chat_history:
            return None
        return self.copilot_chat_history[-1]


class CopilotOutput(BaseModel, ABC):
    """Base class for response events."""

    response_completeness: ResponseCompleteness = Field(
        description=(
            "Indicates whether this is a streaming token (TOKEN) or a complete "
            "response (COMPLETE)"
        ),
    )
    response_category: ResponseCategory = Field(
        description=(
            "The category/source of this response. Each response type has a fixed "
            "category that cannot be changed. "
        ),
        frozen=True,
    )

    @abstractmethod
    def to_sse_event(self) -> ServerSentEvent:
        """Convert to SSE event format."""
        pass


class GeneratedContent(CopilotOutput):
    """Represents generated content from the LLM to be streamed."""

    content: str
    response_category: ResponseCategory = Field(frozen=True)
    response_completeness: ResponseCompleteness = ResponseCompleteness.TOKEN

    def to_sse_event(self) -> ServerSentEvent:
        """Convert to SSE event format."""
        return ServerSentEvent(
            event="copilot_response",
            data={
                "content": self.content,
                "response_category": self.response_category.value,
                "completeness": self.response_completeness.value,
            },
        )


class ReferenceEntry(CopilotOutput):
    """Represents a reference entry with title and url."""

    index: int
    title: str
    url: str
    response_category: ResponseCategory = Field(
        default=ResponseCategory.REFERENCE_ENTRY,
        frozen=True,
    )
    response_completeness: ResponseCompleteness = ResponseCompleteness.COMPLETE

    @model_validator(mode="after")
    def validate_response_category(self) -> "ReferenceEntry":
        """Validate that response_category has the correct default value."""
        if self.response_category != ResponseCategory.REFERENCE_ENTRY:
            raise ValueError(
                f"ReferenceEntry response_category must be "
                f"{ResponseCategory.REFERENCE_ENTRY}, got `{self.response_category}`."
            )
        return self

    def to_sse_event(self) -> ServerSentEvent:
        """Convert to SSE event format."""
        return ServerSentEvent(
            event="copilot_response",
            data={
                "index": self.index,
                "title": self.title,
                "url": self.url,
                "response_category": self.response_category.value,
                "completeness": self.response_completeness.value,
            },
        )


class ReferenceSection(CopilotOutput):
    """Represents a reference section with documentation links."""

    references: list[ReferenceEntry]
    response_category: ResponseCategory = Field(
        default=ResponseCategory.REFERENCE,
        frozen=True,
    )
    response_completeness: ResponseCompleteness = ResponseCompleteness.COMPLETE

    @model_validator(mode="after")
    def validate_response_category(self) -> "ReferenceSection":
        """Validate that response_category has the correct default value."""
        if self.response_category != ResponseCategory.REFERENCE:
            raise ValueError(
                f"ReferenceSection response_category must be "
                f"{ResponseCategory.REFERENCE}, got `{self.response_category}`."
            )
        return self

    def to_sse_event(self) -> ServerSentEvent:
        """Convert to SSE event format."""
        return ServerSentEvent(
            event="copilot_response",
            data={
                "references": [
                    reference.model_dump(include={"index", "title", "url"})
                    for reference in self.references
                ],
                "response_category": self.response_category.value,
                "completeness": self.response_completeness.value,
            },
        )

    def sort_references(self) -> None:
        """Sort references by index value."""
        sorted_references = sorted(
            self.references, key=lambda reference: (0, int(reference.index))
        )

        self.references = sorted_references


class CopilotContext(BaseModel):
    """Model containing the context used by the copilot to generate a response."""

    assistant_logs: str = Field("")
    assistant_files: Dict[str, str] = Field({})
    copilot_chat_history: List["CopilotChatMessage"] = Field([])
    tracker_context: Optional[TrackerContext] = Field(None)

    class Config:
        """Config for LLMBuilderContext."""

        arbitrary_types_allowed = True


class UsageStatistics(BaseModel):
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    model: Optional[str] = None

    def reset(self) -> None:
        """Reset usage statistics to their default values."""
        self.prompt_tokens = None
        self.completion_tokens = None
        self.total_tokens = None
        self.model = None

    def update_from_stream_chunk(self, chunk: ChatCompletionChunk) -> None:
        """Update usage statistics from an OpenAI stream chunk.

        Args:
            chunk: The OpenAI stream chunk containing usage statistics.
        """
        if not (usage := getattr(chunk, "usage", None)):
            return

        self.prompt_tokens = usage.prompt_tokens
        self.completion_tokens = usage.completion_tokens
        self.total_tokens = usage.total_tokens
        self.model = getattr(chunk, "model", None)


class SigningContext(BaseModel):
    secret: Optional[str] = Field(None)
    default_version: str = Field("v1", description="Default signature version")

    @property
    def available(self) -> bool:
        """Signing is enabled if a non-empty secret is present."""
        secret = (self.secret or "").strip()
        return bool(secret)



from __future__ import annotations

import os
import sys
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import structlog
from apscheduler.triggers.cron import CronTrigger
from pydantic import BaseModel, ConfigDict

from rasa.constants import PACKAGE_NAME
from rasa.privacy.constants import (
    ANONYMIZATION_KEY,
    DELETION_KEY,
    KEEP_LEFT_KEY,
    KEEP_RIGHT_KEY,
    PRIVACY_CONFIG_SCHEMA,
    REDACTION_CHAR_KEY,
    SLOT_KEY,
    TRACKER_STORE_SETTINGS,
    USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME,
)
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.io import read_json_file
from rasa.shared.utils.yaml import (
    YamlValidationException,
    validate_data_with_jsonschema,
)

if TYPE_CHECKING:
    from rasa.shared.core.domain import Domain

structlogger = structlog.get_logger(__name__)


class AnonymizationType(Enum):
    """Enum for the anonymization types."""

    REDACT = "redact"
    """Replaces the PII plaintext value with the same character
    for the entire or partial length of the value."""
    MASK = "mask"
    """Replaces the PII plaintext value with the uppercase slot name
    in square brackets, e.g. [CREDIT_CARD_NUMBER]."""


class AnonymizationMethod(BaseModel):
    """Class for configuring the anonymization method."""

    method_type: AnonymizationType
    """The anonymization method to be used."""
    redaction_char: str
    """The character to use for redaction."""
    keep_left: Optional[int] = None
    """The number of characters to be kept intact on the left side."""
    keep_right: Optional[int] = None
    """The number of characters to be kept intact on the right side."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> AnonymizationMethod:
        """Create an AnonymizationMethod object from parsed data."""
        method_type = AnonymizationType(
            data.get("type", AnonymizationType.REDACT.value)
        )
        redaction_char = data.get(REDACTION_CHAR_KEY, "*")
        keep_left = data.get(KEEP_LEFT_KEY)
        keep_right = data.get(KEEP_RIGHT_KEY)

        return cls(
            method_type=method_type,
            redaction_char=redaction_char,
            keep_left=keep_left,
            keep_right=keep_right,
        )


class PrivacyPolicy(BaseModel):
    """Parent class for configuring privacy policies."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    min_after_session_end: int
    """Minimum time in minutes after session end before the policy is executed."""
    cron: CronTrigger
    """Cron trigger for periodic execution of the privacy policy."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> PrivacyPolicy:
        """Create an AnonymizationPolicy object from parsed data."""
        min_after_session_end = data.get("min_after_session_end", 1)
        validate_min_after_session_end(min_after_session_end)

        cron_expression = get_cron_trigger(data.get("cron"))

        return cls(
            min_after_session_end=min_after_session_end,
            cron=cron_expression,
        )


class DeletionPolicy(PrivacyPolicy):
    """Class for configuring periodic deletion in the tracker store."""

    type: str = "deletion"


class AnonymizationPolicy(PrivacyPolicy):
    """Class for configuring periodic anonymization in the tracker store."""

    type: str = "anonymization"


class TrackerStoreSettings(BaseModel):
    """Class for configuring tracker store settings."""

    deletion_policy: Optional[DeletionPolicy] = None
    """The deletion policy to be used."""
    anonymization_policy: Optional[AnonymizationPolicy] = None
    """The anonymization policy to be used."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> TrackerStoreSettings:
        """Create a TrackerStoreSettings object from parsed data."""
        deletion_policy = data.get(DELETION_KEY)
        anonymization_policy = data.get(ANONYMIZATION_KEY)

        deletion_policy = (
            DeletionPolicy.from_dict(deletion_policy) if deletion_policy else None
        )
        anonymization_policy = (
            AnonymizationPolicy.from_dict(anonymization_policy)
            if anonymization_policy
            else None
        )

        validate_policies(deletion_policy, anonymization_policy)

        return cls(
            deletion_policy=deletion_policy,
            anonymization_policy=anonymization_policy,
        )


class PrivacyConfig(BaseModel):
    """Class for configuring PII management."""

    anonymization_rules: Dict[str, AnonymizationMethod]
    """"Mapping of slot names to rules for anonymizing sensitive information."""
    tracker_store_settings: Optional[TrackerStoreSettings] = None
    """The tracker store settings to be used for periodic jobs
    anonymizing and deleting conversation data in the tracker store."""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> PrivacyConfig:
        """Create a PrivacyConfig object from parsed privacy config."""
        # Validate the data against the schema
        validate_privacy_config(data)

        anonymization_rules = {
            rule[SLOT_KEY]: AnonymizationMethod.from_dict(
                rule.get(ANONYMIZATION_KEY, {})
            )
            for rule in data.get("rules", [])
        }

        tracker_store_settings = data.get(TRACKER_STORE_SETTINGS, {})
        tracker_store_settings = (
            TrackerStoreSettings.from_dict(tracker_store_settings)
            if tracker_store_settings
            else None
        )
        return cls(
            anonymization_rules=anonymization_rules,
            tracker_store_settings=tracker_store_settings,
        )


def validate_privacy_config(data: Dict[str, Any]) -> None:
    """Validate the privacy configuration."""
    import importlib_resources

    schema_file = str(
        importlib_resources.files(PACKAGE_NAME).joinpath(PRIVACY_CONFIG_SCHEMA)
    )
    schema_content = read_json_file(schema_file)
    try:
        validate_data_with_jsonschema(data, schema_content)
    except YamlValidationException as exception:
        validation_errors = (
            [error.message for error in exception.validation_errors]
            if exception.validation_errors
            else []
        )
        exception_message = exception.message
        structlogger.error(
            "privacy_config.invalid_privacy_config",
            validation_errors=validation_errors,
            event_info=f"Invalid privacy config: {exception_message}. "
            f"Please check the configuration file.",
        )
        sys.exit(1)


def get_cron_trigger(cron_expression: str) -> CronTrigger:
    """Validate the crontab expression."""
    try:
        cron = CronTrigger.from_crontab(cron_expression)
    except Exception as exc:
        structlogger.error(
            "privacy_config.invalid_cron_expression",
            cron=cron_expression,
        )
        raise RasaException from exc

    return cron


def validate_min_after_session_end(min_after_session_end: int) -> None:
    """Validate the minimum time after session end."""
    try:
        inactivity_period = int(
            os.getenv(USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME, "30")
        )
    except (ValueError, TypeError) as exc:
        raise RasaException(
            f"Invalid value for {USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME} "
            f"env var: {exc}."
        )

    if min_after_session_end < inactivity_period:
        raise RasaException(
            f"Minimum time in minutes after session end must be greater than "
            f"{USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME} env var value."
        )


def validate_policies(
    deletion_policy: Optional[DeletionPolicy],
    anonymization_policy: Optional[AnonymizationPolicy],
) -> None:
    """Validate the deletion and anonymization policies' configurations."""
    if not deletion_policy or not anonymization_policy:
        return None

    if (
        deletion_policy.min_after_session_end
        <= anonymization_policy.min_after_session_end
    ):
        raise RasaException(
            "Minimum time in minutes after session end for deletion policy "
            "must be greater than that of the anonymization policy."
        )

    if deletion_policy.cron.fields == anonymization_policy.cron.fields:
        raise RasaException(
            "Cron expressions for the deletion and anonymization policies "
            "must be different."
        )

    return None


def validate_sensitive_slots(sensitive_slots: List[str], domain: "Domain") -> None:
    """Validate the sensitive slots defined in the privacy config against the domain."""
    all_slot_names = [slot.name for slot in domain.slots]
    all_good = True
    for sensitive_slot in sensitive_slots:
        if sensitive_slot not in all_slot_names:
            structlogger.error(
                "privacy_config.invalid_sensitive_slot",
                sensitive_slot=sensitive_slot,
                event_info="Sensitive slot not found in the domain.",
            )
            all_good = False

    if not all_good:
        raise RasaException(
            "Sensitive slots defined in the privacy config do not match "
            "the slots defined in the domain. Please check the slot names."
        )



class CopilotFinalBufferReached(Exception):
    """Raised when the rolling buffer ends."""

    pass


class CopilotStreamEndedEarly(Exception):
    """Raised when the stream ended early.

    This happens when the stream ended before the max expected special response
    tokens were reached.
    """

    pass


class CopilotStreamError(Exception):
    """Raised when the stream fails."""

    pass


class InvalidCopilotChatHistorySignature(Exception):
    """Raised when the provided history signature does not match."""

    pass


class MissingCopilotChatHistorySignature(Exception):
    """Raised when a required history signature is missing."""

    pass



"""Lakera AI guardrails provider implementation."""

import asyncio
import os
import time
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Dict, Optional

import aiohttp
import structlog

from rasa.builder import config
from rasa.builder.guardrails.constants import (
    LAKERA_API_KEY_ENV_VAR,
    LAKERA_GUARD_ENDPOINT,
    LAKERA_GUARD_RESULTS_ENDPOINT,
)
from rasa.builder.guardrails.exceptions import GuardrailsError
from rasa.builder.guardrails.models import (
    GuardrailResponse,
    LakeraGuardrailRequest,
    LakeraGuardrailResponse,
)

structlogger = structlog.get_logger()


class LakeraAIGuardrails:
    """Guardrails provider using Lakera AI."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        """Initialize Lakera guardrails provider.

        Args:
            severity_threshold: The minimum severity level for the guardrail to flag
                the user request as unsafe.
            api_key: Lakera AI API key.
            base_url: Optional base URL for the API. If not provided, the default
                Lakera API URL (https://api.lakera.ai/v2) will be used.
        """
        self.base_url: str = base_url or config.LAKERA_BASE_URL
        self._api_key: Optional[str] = api_key or os.getenv(LAKERA_API_KEY_ENV_VAR)
        self._session: Optional[aiohttp.ClientSession] = None

    @property
    def guard_endpoint(self) -> str:
        """Get the guard endpoint for the Lakera API."""
        return f"{self.base_url}/{LAKERA_GUARD_ENDPOINT}"

    @property
    def guard_results_endpoint(self) -> str:
        """Get the guard results endpoint for the Lakera API."""
        return f"{self.base_url}/{LAKERA_GUARD_RESULTS_ENDPOINT}"

    @asynccontextmanager
    async def _get_session(self) -> AsyncGenerator[aiohttp.ClientSession, None]:
        """Create a fresh ClientSession, yield it, and always close it."""
        session = aiohttp.ClientSession(headers=self._get_headers())
        structlogger.debug("lakera_guardrails._get_session", base_url=self.base_url)

        try:
            yield session
        except Exception as e:
            structlogger.error("lakera_guardrails.session_error", error=str(e))
            raise
        finally:
            try:
                await session.close()
            except Exception as exc:
                structlogger.warning(
                    "lakera_guardrails.session_close_error",
                    event_info="Failed to close aiohttp client session cleanly.",
                    error=str(exc),
                )

    def _get_headers(self) -> Dict[str, str]:
        """Get the headers for the Lakera API request.

        Returns:
            A dictionary containing the Authorization header with the API key.
        """
        using_proxy = bool(config.HELLO_LLM_PROXY_BASE_URL)

        if using_proxy:
            if not config.RASA_PRO_LICENSE:
                raise GuardrailsError(
                    "HELLO_LLM_PROXY_BASE_URL is set but RASA_PRO_LICENSE is missing. "
                    "Proxy requires a Rasa Pro license token for authentication."
                )
            return {"Authorization": f"Bearer {config.RASA_PRO_LICENSE}"}

        if not self._api_key:
            raise GuardrailsError(
                "LAKERA_API_KEY is missing. Provide it via env LAKERA_API_KEY or "
                "pass api_key= to LakeraAIGuardrails."
            )
        return {"Authorization": f"Bearer {self._api_key}"}

    async def send_request(self, request: LakeraGuardrailRequest) -> GuardrailResponse:
        """Send a request to the Lakera API.

        Args:
            request: The guardrail request to send to the Lakera API.

        Returns:
            GuardrailResponse with the results of the check.

        Raises:
            GuardrailsError: If the request times out or returns a non-200 status code.
            Exception: If the request fails for any other reason.
        """
        start_time = time.time()
        try:
            async with self._get_session() as session:
                raw_response = await self._send_http_request(session, request)
                response = LakeraGuardrailResponse.from_raw_response(
                    raw_response,
                    hello_rasa_user_id=request.hello_rasa_user_id,
                    hello_rasa_project_id=request.hello_rasa_project_id,
                )
                processing_time_ms = (time.time() - start_time) * 1000
                response.processing_time_ms = processing_time_ms
                return response

        # Propagate the GuardrailsError if it occurs.
        except GuardrailsError as e:
            raise e

        except asyncio.TimeoutError:
            message = "Lakera API request timed out."
            structlogger.error(
                "lakera_guardrails.send_request.timeout_error",
                event_info=message,
                processing_time_ms=(time.time() - start_time) * 1000,
            )
            raise GuardrailsError(message)

        # Propagate the unexpected exceptions.
        except Exception as e:
            message = "Lakera API request failed."
            structlogger.error(
                "lakera_guardrails.send_request.unexpected_error",
                event_info="Lakera API request failed.",
                request=request,
                error=e,
                processing_time_ms=(time.time() - start_time) * 1000,
            )
            raise e

    async def _send_http_request(
        self, session: aiohttp.ClientSession, request: LakeraGuardrailRequest
    ) -> Dict[str, Any]:
        """Make an HTTP request to the Lakera API.

        Args:
            session: The aiohttp session to use for the request.
            request: The guardrail request to send.

        Returns:
            The raw JSON response from the API.

        Raises:
            GuardrailsError: If the request fails or response parsing fails.
        """
        # Log the request details for debugging
        json_payload = request.to_json_payload()
        structlogger.debug(
            "lakera_guardrails.send_request.request",
            url=self.guard_endpoint,
            method="POST",
            request_body=json_payload,
        )

        async with session.post(
            self.guard_endpoint,
            json=json_payload,
        ) as client_response:
            # Check if the response is successful. If not, raise an error.
            if client_response.status >= 400:
                error_text = await client_response.text()
                message = (
                    f"Lakera API request failed with status "
                    f"`{client_response.status}`. Error: "
                    f"`{error_text}`."
                )
                structlogger.error(
                    "lakera_guardrails.send_request.http_error",
                    event_info=message,
                    url=self.guard_endpoint,
                    status=client_response.status,
                    error=error_text,
                    request_body=json_payload,
                )
                raise GuardrailsError(message)

            # Parse the response as a dictionary.
            raw_response = await client_response.json()
            structlogger.debug(
                "lakera_guardrails.send_request.response",
                response_body=raw_response,
            )
            return raw_response



import asyncio
from functools import lru_cache
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple

import structlog

from rasa.builder.config import (
    ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
    COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
)
from rasa.builder.copilot.constants import ROLE_COPILOT, ROLE_USER
from rasa.builder.copilot.copilot_response_handler import CopilotResponseHandler
from rasa.builder.copilot.models import (
    CopilotChatMessage,
    CopilotContext,
    GeneratedContent,
    ResponseCategory,
)
from rasa.builder.guardrails.models import (
    GuardrailRequestKey,
    GuardrailResponse,
    LakeraGuardrailRequest,
)
from rasa.builder.llm_service import llm_service
from rasa.builder.shared.tracker_context import (
    AssistantConversationTurn,
    TrackerContext,
)

if TYPE_CHECKING:
    from rasa.builder.guardrails.models import GuardrailType


structlogger = structlog.get_logger()


def map_lakera_detector_type_to_guardrail_type(
    lakera_detector_type: str,
) -> Optional["GuardrailType"]:
    """Map a Lakera detector type to a Rasa guardrail type."""
    from rasa.builder.guardrails.models import GuardrailType

    # Check for exact matches first
    LAKERA_DETECTOR_TYPES_2_RASA_GUARDRAIL_TYPES_MAPPING = {
        "prompt_attack": GuardrailType.PROMPT_ATTACK,
        "unknown_links": GuardrailType.MALICIOUS_LINKS,
        "custom": GuardrailType.CUSTOM,
    }

    # Check for exact match first
    if lakera_detector_type in LAKERA_DETECTOR_TYPES_2_RASA_GUARDRAIL_TYPES_MAPPING:
        return LAKERA_DETECTOR_TYPES_2_RASA_GUARDRAIL_TYPES_MAPPING[
            lakera_detector_type
        ]

    # Check for subtypes that start with specific prefixes
    # https://docs.lakera.ai/docs/policies/self-hosted-policies#detectors-section
    if lakera_detector_type.startswith("moderated_content"):
        return GuardrailType.CONTENT_VIOLATION
    if lakera_detector_type.startswith("pii"):
        return GuardrailType.DATA_LEAKAGE

    # If no match found, return OTHER
    return GuardrailType.OTHER


@lru_cache(maxsize=512)
def _schedule_guardrails_check(
    user_text: str,
    hello_rasa_user_id: str,
    hello_rasa_project_id: str,
    lakera_project_id: str,
) -> "asyncio.Task[GuardrailResponse]":
    """Return a cached asyncio.Task that resolves to Lakera's response.

    Args:
        user_text: The user message text to check for policy violations.
        hello_rasa_user_id: The user ID for the conversation.
        hello_rasa_project_id: The project ID for the conversation.
        lakera_project_id: The Lakera project ID to use for this check.

    Returns:
        An asyncio Task that resolves to a GuardrailResponse.
    """
    structlogger.debug("guardrails.cache_miss", text=user_text)

    loop = asyncio.get_running_loop()
    request = LakeraGuardrailRequest(
        lakera_project_id=lakera_project_id,
        hello_rasa_user_id=hello_rasa_user_id,
        hello_rasa_project_id=hello_rasa_project_id,
        messages=[{"role": ROLE_USER, "content": user_text}],
    )

    return loop.create_task(llm_service.guardrails.send_request(request))


async def _detect_flagged_user_indices(
    items: List[Tuple[int, str]],
    *,
    hello_rasa_user_id: Optional[str],
    hello_rasa_project_id: Optional[str],
    lakera_project_id: str,
    log_prefix: str,
) -> Set[int]:
    """Run guardrail checks for provided (index, user_text) pairs.

    Args:
        items: List of tuples containing (index, user_text) to check.
        hello_rasa_user_id: The user ID for the conversation.
        hello_rasa_project_id: The project ID for the conversation.
        lakera_project_id: The Lakera project ID to use for this check.
        log_prefix: Prefix for logging messages.

    Returns:
        A set of indices that were flagged by the guardrails.
    """
    if not items:
        return set()

    # 1) Group indices by logical request key (hashable by value)
    indices_by_key: Dict[GuardrailRequestKey, List[int]] = {}
    for idx, text in items:
        key = GuardrailRequestKey(
            user_text=(text or "").strip(),
            hello_rasa_user_id=hello_rasa_user_id or "",
            hello_rasa_project_id=hello_rasa_project_id or "",
            lakera_project_id=lakera_project_id,
        )
        if not key.user_text:
            continue
        indices_by_key.setdefault(key, []).append(idx)

    if not indices_by_key:
        return set()

    # 2) Create one task per logical key
    tasks_by_key: Dict[GuardrailRequestKey, asyncio.Task[GuardrailResponse]] = {}
    for key in indices_by_key:
        tasks_by_key[key] = _schedule_guardrails_check(
            user_text=key.user_text,
            hello_rasa_user_id=key.hello_rasa_user_id,
            hello_rasa_project_id=key.hello_rasa_project_id,
            lakera_project_id=key.lakera_project_id,
        )

    # 3) Await unique tasks once
    keys = list(tasks_by_key.keys())
    tasks = [tasks_by_key[k] for k in keys]
    responses = await asyncio.gather(*tasks, return_exceptions=True)

    # 4) Map results back to all corresponding indices
    flagged: Set[int] = set()
    for key, response in zip(keys, responses):
        if isinstance(response, Exception):
            structlogger.warning(f"{log_prefix}.request_failed", error=str(response))
            continue
        if response.flagged:
            flagged.update(indices_by_key.get(key, []))

    return flagged


async def check_assistant_chat_for_policy_violations(
    tracker_context: TrackerContext,
    hello_rasa_user_id: Optional[str],
    hello_rasa_project_id: Optional[str],
) -> TrackerContext:
    """Return a sanitised TrackerContext with unsafe turns removed.

    Only user messages are moderated – assistant messages are assumed safe.
    LRU cache is used, so each unique user text is checked once.

    Args:
        tracker_context: The TrackerContext containing conversation turns.
        hello_rasa_user_id: The user ID for the conversation.
        hello_rasa_project_id: The project ID for the conversation.

    Returns:
        TrackerContext with unsafe turns removed.
    """
    # Collect (turn_index, user_text) for all turns with a user message
    items: List[Tuple[int, str]] = []
    for idx, turn in enumerate(tracker_context.conversation_turns):
        user_message = turn.user_message
        if not user_message:
            continue

        text = (user_message.text or "").strip()
        if not text:
            continue

        items.append((idx, text))

    flagged_turns = await _detect_flagged_user_indices(
        items,
        hello_rasa_user_id=hello_rasa_user_id,
        hello_rasa_project_id=hello_rasa_project_id,
        lakera_project_id=ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
        log_prefix="assistant_guardrails",
    )

    if not flagged_turns:
        return tracker_context

    structlogger.info(
        "assistant_guardrails.turns_flagged",
        count=len(flagged_turns),
        turn_indices=sorted(flagged_turns),
    )

    # Build a filtered TrackerContext
    safe_turns: List[AssistantConversationTurn] = [
        turn
        for idx, turn in enumerate(tracker_context.conversation_turns)
        if idx not in flagged_turns
    ]

    new_tracker_context = tracker_context.copy(deep=True)
    new_tracker_context.conversation_turns = safe_turns
    return new_tracker_context


def _annotate_flagged_user_messages(
    history: List[CopilotChatMessage], flagged_user_indices: Set[int]
) -> None:
    """Mark flagged user messages in-place on the original history.

    Args:
        history: The copilot chat history containing messages.
        flagged_user_indices: Set of indices of user messages that were flagged.
    """
    if not flagged_user_indices:
        return

    total = len(history)
    for uidx in flagged_user_indices:
        if 0 <= uidx < total and history[uidx].role == ROLE_USER:
            history[
                uidx
            ].response_category = ResponseCategory.GUARDRAILS_POLICY_VIOLATION


async def check_copilot_chat_for_policy_violations(
    context: CopilotContext,
    hello_rasa_user_id: Optional[str],
    hello_rasa_project_id: Optional[str],
) -> Optional[GeneratedContent]:
    """Check the copilot chat history for guardrail policy violations.

    Only user messages are moderated – assistant messages are assumed safe.
    LRU cache is used, so each unique user text is checked once.

    Args:
        context: The CopilotContext containing the copilot chat history.
        hello_rasa_user_id: The user ID for the conversation.
        hello_rasa_project_id: The project ID for the conversation.

    Returns:
        Returns a default violation response if the system flags any user message,
        otherwise return None.
    """
    history = context.copilot_chat_history

    # Collect (index, text) for user messages; skip ones already marked as violations
    items: List[Tuple[int, str]] = []
    for idx, message in enumerate(history):
        if message.response_category == ResponseCategory.GUARDRAILS_POLICY_VIOLATION:
            continue
        if message.role != ROLE_USER:
            continue
        formatted_message = message.to_openai_format()
        text = (formatted_message.get("content") or "").strip()
        if not text:
            continue
        items.append((idx, text))

    flagged_user_indices = await _detect_flagged_user_indices(
        items,
        hello_rasa_user_id=hello_rasa_user_id,
        hello_rasa_project_id=hello_rasa_project_id,
        lakera_project_id=COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
        log_prefix="copilot_guardrails",
    )

    _annotate_flagged_user_messages(history, flagged_user_indices)

    if not flagged_user_indices:
        return None

    # Identify the latest user message index in the current request
    last_user_idx: Optional[int] = None
    for i in range(len(history) - 1, -1, -1):
        if getattr(history[i], "role", None) == ROLE_USER:
            last_user_idx = i
            break

    # Remove flagged user messages and their next copilot messages
    indices_to_remove: Set[int] = set()
    total = len(history)
    for uidx in flagged_user_indices:
        indices_to_remove.add(uidx)
        next_idx = uidx + 1
        if (
            next_idx < total
            and getattr(history[next_idx], "role", None) == ROLE_COPILOT
        ):
            indices_to_remove.add(next_idx)

    # Apply sanitization
    filtered_history = [
        msg for i, msg in enumerate(history) if i not in indices_to_remove
    ]
    if len(filtered_history) != len(history):
        structlogger.info(
            "copilot_guardrails.history_sanitized",
            removed_indices=sorted(indices_to_remove),
            removed_messages=len(history) - len(filtered_history),
            kept_messages=len(filtered_history),
        )
        context.copilot_chat_history = filtered_history

    # Block only if the latest user message in this request was flagged
    if last_user_idx is not None and last_user_idx in flagged_user_indices:
        return CopilotResponseHandler.respond_to_guardrail_policy_violations()

    # Otherwise proceed (following messages are respected)
    return None



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionRemovePayee(Action):
    def name(self) -> Text:
        return "action_remove_payee"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")
        payee_name = tracker.get_slot("payee_name")

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Check if payee exists
        payee = db.get_payee_by_name_and_user(payee_name, int(user["id"]))
        if not payee:
            dispatcher.utter_message(
                text=f"I'm sorry, but I couldn't find a payee named '{payee_name}'"
            )
            return [SlotSet("payee_removed", False)]

        # Remove the payee
        success = db.remove_payee(payee_name, int(user["id"]))

        if success:
            dispatcher.utter_message(
                text=f"Payee {payee_name} has been removed successfully."
            )
            return [SlotSet("payee_removed", True)]
        else:
            dispatcher.utter_message(text="Failed to remove payee. Please try again.")
            return [SlotSet("payee_removed", False)]



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.database import Database


class ActionCheckPayeeExistence(Action):
    def name(self) -> Text:
        return "action_check_payee_existence"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        name = tracker.get_slot("name")
        payee_name = tracker.get_slot("payee_name")

        db = Database()

        # Get user information
        user = db.get_user_by_name(name)
        if not user:
            dispatcher.utter_message(text="User not found.")
            return []

        # Check if payee exists
        payee = db.get_payee_by_name_and_user(payee_name, int(user["id"]))

        if payee:
            return [SlotSet("payee_exists", True)]

        dispatcher.utter_message(
            text=f"{payee_name} is not an authorised payee. Let's add them!"
        )
        return [SlotSet("payee_exists", False)]



# Account actions
from actions.accounts.action_ask_account import ActionAskAccount
from actions.accounts.action_check_balance import ActionCheckBalance

# General actions
from actions.action_session_start import ActionSessionStart

# Card actions
from actions.cards.action_ask_card import ActionAskCard
from actions.cards.action_check_card_existence import ActionCheckCardExistence
from actions.cards.action_update_card_status import ActionUpdateCardStatus

# Transfer actions
from actions.transfers.action_add_payee import ActionAddPayee
from actions.transfers.action_ask_account_from import ActionAskAccountFrom
from actions.transfers.action_check_payee_existence import ActionCheckPayeeExistence
from actions.transfers.action_check_sufficient_funds import ActionCheckSufficientFunds
from actions.transfers.action_list_payees import ActionListPayees
from actions.transfers.action_process_immediate_payment import (
    ActionProcessImmediatePayment,
)
from actions.transfers.action_remove_payee import ActionRemovePayee
from actions.transfers.action_schedule_payment import ActionSchedulePayment
from actions.transfers.action_validate_payment_date import ActionValidatePaymentDate

__all__ = [
    # General actions
    "ActionSessionStart",
    # Account actions
    "ActionAskAccount",
    "ActionCheckBalance",
    # Card actions
    "ActionAskCard",
    "ActionCheckCardExistence",
    "ActionUpdateCardStatus",
    # Transfer actions
    "ActionAddPayee",
    "ActionAskAccountFrom",
    "ActionCheckPayeeExistence",
    "ActionCheckSufficientFunds",
    "ActionListPayees",
    "ActionProcessImmediatePayment",
    "ActionRemovePayee",
    "ActionSchedulePayment",
    "ActionValidatePaymentDate",
]



import argparse
from typing import List, Text

from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_config_param,
    add_data_param,
    add_domain_param,
    add_endpoint_param,
)
from rasa.shared.constants import (
    DEFAULT_DOMAIN_PATH,
    DEFAULT_ENDPOINTS_PATH,
)
from rasa.studio.push import (
    handle_push,
    handle_push_config,
    handle_push_endpoints,
)


def add_subparser(
    subparsers: SubParsersAction,
    parents: List[argparse.ArgumentParser],
    domain: Text = DEFAULT_DOMAIN_PATH,
) -> None:
    """Register the `rasa studio push` command and its sub-commands.

    Args:
        subparsers: The subparsers to add the command to.
        parents: Parent parsers, needed to ensure tree structure in argparse
        domain: Path to the assistant's domain file.
    """
    push_parser = subparsers.add_parser(
        "push",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Push assistant data to Rasa Studio.",
    )
    push_subparsers = push_parser.add_subparsers()

    # Arguments for pushing configuration only
    config_parser = push_subparsers.add_parser(
        "config",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Push only the assistant configuration file.",
    )
    add_config_param(config_parser)
    config_parser.set_defaults(func=handle_push_config)

    # Arguments for pushing endpoints only
    endpoints_parser = push_subparsers.add_parser(
        "endpoints",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Push ONLY the endpoints configuration file.",
    )
    add_endpoint_param(
        endpoints_parser,
        "Configuration file for the model endpoints.",
        default=DEFAULT_ENDPOINTS_PATH,
    )
    endpoints_parser.set_defaults(func=handle_push_endpoints)

    # Arguments for pushing the whole assistant data
    add_domain_param(push_parser, domain)
    add_data_param(push_parser)
    add_config_param(push_parser)
    add_endpoint_param(
        push_parser,
        "Configuration file for the model endpoints.",
        default=DEFAULT_ENDPOINTS_PATH,
    )
    push_parser.set_defaults(func=handle_push)



from typing import Dict, List, Text, Union


def add_item_to_lookup_tables(
    title: Text,
    item: Text,
    existing_lookup_tables: List[Dict[Text, Union[Text, List[Text]]]],
) -> None:
    """Add an item to a list of existing lookup tables.

    Takes a list of lookup table dictionaries. Finds the one associated
    with the current lookup, then adds the item to the list.

    Args:
        title: Name of the lookup item.
        item: The lookup item.
        existing_lookup_tables: Existing lookup items that will be extended.

    Raises:
        TypeError: in case we're trying to add a lookup table element to a file.
                   This is an internal error that is indicative of a parsing error.
    """
    matches = [table for table in existing_lookup_tables if table["name"] == title]
    if not matches:
        existing_lookup_tables.append({"name": title, "elements": [item]})
    else:
        elements = matches[0]["elements"]
        if not isinstance(elements, list):
            raise TypeError("Cannot add a lookup table element to an unloaded file.")
        elements.append(item)



from typing import Any, Dict, List, Text

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher


class ActionSchedulePayment(Action):
    def name(self) -> Text:
        return "action_schedule_payment"

    def run(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict[Text, Any]]:
        # TODO: Should this add the payment to a scheduled_payment table?
        return [SlotSet("payment_scheduled", True)]
